 !{\src2tex{textfont=tt}}
!!****m* ABINIT2XML/abinit2xml
!! NAME
!! abinit2xml
!!
!! FUNCTION
!! Use this program to convert a PAW atomic dataset
!!  from ABINIT proprietary format
!!  to XML format
!!  See PAW-XML specification: https://esl.cecam.org/Paw-xml
!!
!! COPYRIGHT
!! Copyright (C) 2014-2018 ABINIT group (FJ, MT)
!! This file is distributed under the terms of the GNU General Public License
!! See http://www.gnu.org/copyleft/gpl.txt
!!
!! NOTES
!! The present main routine drives the following operations :
!! 1) Read data from PAW file in (legacy) ABINIT proprietary format
!! 2) Write an XML-formatted output file
!! Units: Energies=Hartree, Lengths=Bohr

!! SOURCE

 program abinit2xml

 implicit none


!================================================================================
! USEFUL CONSTANTS AND PARAMETERS
!================================================================================

!Nb of bytes related to default double-precision real subtypes
 integer, parameter :: dp=kind(1.0d0)
 integer, parameter :: dpc=kind((1.0d0,1.0d0))  ! Complex should not be used presently

!Default string lengths
 integer, parameter :: fnlen=132    ! maximum length of file name variables
 integer, parameter :: strlen=32000 ! maximum length of input string

!UNIX unit numbers
 integer, parameter :: std_in=5,std_out=6

!Real constants
 real(dp), parameter :: zero=0._dp
 real(dp), parameter :: one=1._dp
 real(dp), parameter :: two=2._dp
 real(dp), parameter :: three=3._dp
 real(dp), parameter :: four=4._dp
 real(dp), parameter :: half=0.50_dp
 real(dp), parameter :: third=one/three
 real(dp), parameter :: quarter=0.25_dp
 real(dp), parameter :: pi=3.141592653589793238462643383279502884197_dp
 real(dp), parameter :: two_pi=2._dp*pi
 real(dp), parameter :: four_pi=4._dp*pi

!Real precision
 real(dp), parameter :: greatest_real = huge(one)
 real(dp), parameter :: smallest_real = -greatest_real
 real(dp), parameter :: tol6= 0.000001_dp
 real(dp), parameter :: tol8= 0.00000001_dp
 real(dp), parameter :: tol10=0.0000000001_dp
 real(dp), parameter :: tol11=0.00000000001_dp
 real(dp), parameter :: tol12=0.000000000001_dp
 real(dp), parameter :: tol14=0.00000000000001_dp
 real(dp), parameter :: tol16=0.0000000000000001_dp

!Real physical constants
 real(dp), parameter :: Bohr_Ang=0.5291772108_dp    ! 1 Bohr, in Angstrom
 real(dp), parameter :: Ha_eV=27.2113845_dp ! 1 Hartree, in eV

!Character constants
 character(len=1), parameter :: ch10 = char(10)


!================================================================================
! STRUCTURED DATATYPES
!================================================================================

!PAW dataset "header" for ABINIT
 type pshead_type
  integer :: basis_size      ! Number of elements for the paw nl basis
  integer :: core_size       ! Number of core states
  integer :: core_meshsz     ! Dimension of radial mesh for core density
                             ! Must be greater or equal than sph_meshsz
  integer :: corewf_meshsz   ! Dimension of radial mesh for core wave functions
  integer :: creatorid       ! ID of psp generator (here creatorID=1 !)
  integer :: hat_meshsz      ! Dimension of radial mesh for hat density
  integer :: l_size          ! Maximum value of l+1 leading to a non zero Gaunt coef
  integer :: lambda          ! Lambda in gaussian type g(r)
  integer :: lmax            ! Maximum value of l for valence states
  integer :: lmax_core       ! Maximum value of l for core states
  integer :: lmn_size        ! Number of elements for the paw basis
  integer :: lmn_size_core   ! Number of lmn elements for the core
  integer :: lmn2_size       ! lmn2_size=lmn_size*(lmn_size+1)/2
  integer :: max_meshsz      ! max. dimension for radial meshes
  integer :: mesh_type       ! Flag: 1 if all meshes are regular grids
                             !       2 if all meshes are log. grids
  integer :: nmesh          ! max. number of radial meshes
  integer :: prj_meshsz      ! Dimension of radial mesh for tproj
  integer :: prj_msz_max     ! Maximal dimension of radial mesh for tproj
  integer :: pspcod          ! Psp code number for Abinit (here PAW->pspcod=7 !)
  integer :: pspversion      ! Version of the PAW format for Abinit
  integer :: pspxc_abinit    ! Abinit code number for the XC functionnal
  integer :: shape_type      ! Shape function type
                             ! shape_type=1 ; g(r)=sinc2
                             ! shape_type=2 ; g(r)=exp
                             ! shape_type=2 ; g(r)=bessel
                             ! shape_type=-1; g(r)=tabulated
  integer :: sph_meshsz      ! Dimension of radial mesh corresponding to PAW radius
  integer :: vale_meshsz     ! Dimension of radial mesh for pseudo valence density
  integer :: vloc_meshsz     ! Dimension of radial mesh for vloc=vhtnzc (with hat in XC)
                             ! For r>r(vloc_meshsz), we assume vloz=-Z/r
  integer :: vlocopt         ! Option for the format of Vloc
                             !  0: Vloc is Vbare (Blochl formulation)
                             !  1: Vloc is VH(tnzc) (Kresse formulation)
                             !  2: Vloc is VH(tnzc) (Kresse formulation without compensation in XC)
  integer :: wav_meshsz      ! Dimension of radial mesh for wave functions (phi, tphi)
  real(dp) :: atomic_charge  ! Total atomic charge
  real(dp) :: core_charge    ! Core charge
  real(dp) :: log_step       ! Step corresponding to exponential term in logarith. radial mesh
  real(dp) :: rad_step       ! Step corresponding to radial mesh
  real(dp) :: rc_hat         ! Radius used in sinus shape function
  real(dp) :: rc_sph         ! Default PAW sphere radius
  real(dp) :: sigma          ! Sigma for gaussian shape function
  character*(fnlen) :: title ! Title for pseudopotential
  character (len=8) :: pspdat          ! Date of creation
  integer, pointer :: orbitals(:)      ! orbitals(basis_size) (l quantum number for each basis function)
  integer, pointer :: orbitals_core(:) ! orbitals_core(basis_size) (l quantum number for each core WF)
  real(dp), pointer :: occ(:)          ! occ(basis_size) (occupation for each basis function)
 end type pshead_type

!ABINIT PAW dataset (except header)
 type pawps_type
  real(dp), allocatable :: coredens(:)  ! Core density
  real(dp), allocatable :: tcoredens(:) ! Pseudo core density
  real(dp), allocatable :: tvaledens(:) ! Pseudo valence density multiplied
  real(dp), allocatable :: phi(:,:)     ! PAW atomic partial waves
  real(dp), allocatable :: tphi(:,:)    ! PAW atomic pseudo partial waves
  real(dp), allocatable :: tproj(:,:)   ! PAW projectors for each basis function
  real(dp), allocatable :: dij0(:)      ! Part of the Dij term frozen during SCF
  real(dp), allocatable :: kij(:)       ! Kinetic overlap operator
  real(dp), allocatable :: rhoij0(:)    ! Atomic initialization of rhoij
  real(dp), allocatable :: shpf(:,:)    ! l-dependent shapefunction
  real(dp), allocatable :: vbare(:)     ! "zero" local potential
  real(dp), allocatable :: vhtnzc(:)    ! Hartree potential of the ps-density of (nucleus+core_el)
 end type pawps_type

!Grid definitions
 type pawrad_type
  integer :: islog                    ! 0 if the radial grid is a linear grid, 1 if it is a logarithmic grid
  integer :: mesh_size                ! Dimension of max. radial mesh (max. of all mesh sizes)
  integer :: mesh_type                ! Type of radial mesh (log, lin, etc..)
  integer :: int_meshsz               ! Mesh size used in integrals computation
  real(dp) :: rstep                   ! Step corresponding to radial mesh
  real(dp) :: lstep                   ! Step corresponding to exponential term (logarithmic mesh)
  real(dp) :: stepint                 ! Radial step used for integration by the trapeze method
  real(dp), allocatable :: rad(:)     ! rad(max_meshsz)  Coordinates of all the pts of the radial grid
  real(dp), allocatable :: radfact(:) ! radfact(max_meshsz) used to compute radial integrals on generalized grid 
  real(dp), allocatable :: simfact(:) ! Simpson factors
 end type pawrad_type


!================================================================================
! GLOBAL PARAMETERS
!================================================================================

 character(len=10) :: progname='abinit2xml'
 character(len=5) :: version='1.1.0'

!================================================================================
! MAIN PROGRAM VARIABLES
!================================================================================

 integer, parameter :: nmesh_max=6
 integer:: mesh_indx(nmesh_max)
 logical :: file_exists
 character(len=fnlen) :: abinit_file,xml_file
 type(pshead_type):: pshead
 type(pawrad_type):: radmesh(nmesh_max)
 type(pawps_type) :: pawps


!================================================================================
! MAIN PROGRAM
!================================================================================

!Get command line argument
 if (iargc()>=1) then
   call getarg(1,abinit_file)
   if (trim(abinit_file)=='--help'.or.trim(abinit_file)=='-h'.or.iargc()>1) then
     print *,'Usage: '//progname//' [input_file] [--help] [-h]'
     stop
   end if
 end if
 xml_file=trim(abinit_file)//'.xml'

!Print title
 print *,progname//' v'//version//':'
 print *,' Convert a ABINIT-formatted PAW atomic dataset into a XML-formatted dataset...'

!Check input file
 inquire(file=trim(abinit_file),exist=file_exists)
 if (.not.file_exists) then
   print *,'  File '//trim(abinit_file)//' does not exist!'
   stop
 end if

!Run
 call rdpawps(abinit_file,mesh_indx,pawps,pshead,radmesh)
 call wrxmlps(xml_file,mesh_indx,pawps,pshead,radmesh)
 print *,' ...File '//trim(xml_file)//' created'

 CONTAINS


!================================================================================
! ROUTINE: READING OF THE ABINIT-FORMATTED FILE
!================================================================================

!! FUNCTION
!!  Read a PAW data file at the abinit format
!!
!! INPUTS
!!  abinit_file_toread=name of the input ABINIT-formatted PAW dataset
!!
!! OUTPUT
!!  mesh_indx(:)= indexes the different meshes
!!  pshead
!!    %atomic_charge= Total atomic charge
!!    %basis_size= Number of elements for the paw nl basis
!!    %core_charge= Core charge
!!    %vlocopt=Option for the format of Vloc
!!              0: Vloc is Vbare (Blochl formulation)
!!              1: Vloc is VH(tnzc) (Kresse formulation)
!!              2: Vloc is VH(tnzc) (Kresse formulation without compensation in XC)
!!    %lambda= Lambda in gaussian type g(r)
!!    %lmn_size= Number of elements for the paw basis
!!    %lmn2_size=lmn_size*(lmn_size+1)/2
!!    %l_size = Maximum value of l+1 leading to a non zero Gaunt coef
!!    %max_meshsz   = max. number of radial meshes
!!    %orbitals(basis_size)= Quantum number l for each basis function
!!    %pspxc_abinit= ABINIT code number for the exchange-correlation
!!    %pspversion= Version of the PAW format for Abinit
!!    %rc_hat= radius for shape function
!!    %rc_sph= Default PAW sphere radius
!!    %shape_type= Shape function type
!!    %sigma= Sigma for gaussian type g(r)
!!    %title= Title for pseudopotential
!! pawps
!!  %coredens(:)  ! Core density
!!  %tcoredens(:) ! Pseudo core density
!!  %tvaledens(:) ! Pseudo valence density multiplied
!!  %phi(:,:)     ! PAW atomic partial waves
!!  %tphi(:,:)    ! PAW atomic pseudo partial waves
!!  %tproj(:,:)   ! PAW projectors for each basis function
!!  %dij0(:)      ! Part of the Dij term frozen during SCF
!!  %kij(:)       ! Kinetic overlap operator
!!  %rhoij0(:)    ! Atomic initialization of rhoij
!!  %shpf(:,:)    ! l-dependent shapefunction
!!  %vhtnzc(:)    ! Hartree potential of the ps-density of (nucleus+core_el)
!! radmesh(:)= all the meshes

 subroutine rdpawps(abinit_file_toread,mesh_indx,pawps,pshead,radmesh)
 
!Arguments
 integer,intent(out) :: mesh_indx(nmesh_max)
 character(len=fnlen),intent(in) :: abinit_file_toread
 type(pshead_type),intent(inout) :: pshead
 type(pawrad_type),intent(inout) :: radmesh(nmesh_max)
 type(pawps_type),intent(inout)  :: pawps

!Local variables
 integer, parameter :: in_unit=10
 integer :: creatorid,ib,icoremesh,ii,jj,il,imainmesh,imsh,iprojmesh
 integer :: ir,iread1,iread2,ishpfmesh,ivalemesh,ivlocmesh
 integer :: lmax,lloc,mmax,nmesh,pspcod,pspversion,usetvale
 real(dp) :: r2well,rc,rread1,rread2,rshp0,zion
 character(len=80) :: pspline
 character(len=500) :: message,submessage
 character :: blank=' ',numb=' '

!------------------------------------------------------------------

 open(in_unit,file=trim(abinit_file_toread),form='formatted')

!------------------------------------------------------------------
!--------------- Header

!Line 1: read title
 read (in_unit,'(a)') pshead%title

!Line 2: read charges, XC, sizes, ...
 read (in_unit,*) pshead%atomic_charge,zion,pshead%pspdat
 pshead%core_charge=pshead%atomic_charge-zion
 read (in_unit,*) pspcod,pshead%pspxc_abinit,lmax,lloc,mmax,r2well

!Line 3: read psp version in line 4 of the header
 pspversion=1
 read (in_unit,'(a80)') pspline;pspline=adjustl(pspline)
 if (pspline(1:3)=="paw".or.pspline(1:3)=="PAW") &
& read(unit=pspline(4:80),fmt=*) pspversion
 pshead%pspversion=pspversion
 if (pspversion<1.or.pspversion>5) then
   write(std_out, '(a,i2,a,a,a)' )&
&   '  This version of PAW psp file (',pspversion,') is not compatible with',ch10,&
&   '  current version of Abinit.'
   stop
 end if

!Lines 4-11: def. of the projectors, radial meshes, shape function, ...
 if (pspversion==1) then
! VERSION 1 (Abinit<v4.3)
   read (unit=pspline,fmt=*) pshead%basis_size,pshead%lmn_size
   allocate(pshead%orbitals(pshead%basis_size))
   read(in_unit,*) (pshead%orbitals(ib), ib=1,pshead%basis_size)
   pshead%l_size=2*maxval(pshead%orbitals)+1
   nmesh=3
   read(in_unit,'(a80)') pspline
   radmesh(1)%lstep=zero
   read(unit=pspline,fmt=*,err=10,end=10) radmesh(1)%mesh_type,&
&   radmesh(1)%rstep,radmesh(1)%lstep
  10 read(in_unit,*) pshead%rc_sph
   read(in_unit,*) radmesh(1)%mesh_size,radmesh(2)%mesh_size,&
&   radmesh(3)%mesh_size
   read(in_unit,'(a80)') pspline
   pshead%lambda=-1;pshead%sigma=1.d99
   read(unit=pspline,fmt=*,err=11,end=11) pshead%shape_type,&
&   pshead%lambda,pshead%sigma
   11 read(in_unit,*) creatorid
   if (pshead%shape_type==3) pshead%shape_type=-1
   radmesh(2)%mesh_type=radmesh(1)%mesh_type
   radmesh(3)%mesh_type=radmesh(1)%mesh_type
   radmesh(2)%rstep=radmesh(1)%rstep
   radmesh(3)%rstep=radmesh(1)%rstep
   radmesh(2)%lstep=radmesh(1)%lstep
   radmesh(3)%lstep=radmesh(1)%lstep
 else
! VERSION 2+ (Abinit>=v4.3)
   read (unit=pspline(5:80),fmt=*) creatorid
   read (in_unit,*) pshead%basis_size,pshead%lmn_size
   allocate(pshead%orbitals(pshead%basis_size))
   read(in_unit,*) (pshead%orbitals(ib), ib=1,pshead%basis_size)
   pshead%l_size=2*maxval(pshead%orbitals)+1
   read(in_unit,*) nmesh
   if (nmesh>10) then
     write(std_out, '(3a)' )&
&       '  Index of mesh out of range!',ch10,&
&       '  Action : check your pseudopotential file.'
     stop
   end if
   do imsh=1,nmesh
     rread2=zero
     read(in_unit,'(a80)') pspline
     read(unit=pspline,fmt=*,err=20,end=20) ii,iread1,iread2,rread1,rread2
     20 continue
     if (ii<=nmesh) then
       radmesh(ii)%mesh_type=iread1
       radmesh(ii)%mesh_size=iread2
       radmesh(ii)%rstep=rread1
       radmesh(ii)%lstep=rread2
     else
       write(std_out, '(3a)' )&
&       '  Index of mesh out of range!',ch10,&
&       '  Action : check your pseudopotential file.'
       stop
     end if
   end do
   read(in_unit,*) pshead%rc_sph
   read(in_unit,'(a80)') pspline
   read(unit=pspline,fmt=*) pshead%shape_type
   pshead%lambda=-1;pshead%sigma=1.d99
 end if
 pshead%nmesh=nmesh
 do ii=1,nmesh
   call init_radmesh(radmesh(ii),0.d0)
 end do
 pshead%lmn2_size=pshead%lmn_size*(pshead%lmn_size+1)/2
 if (pspversion==2) then
!  Shapefunction parameters for Abinit 4.3...4.5
   if (pshead%shape_type==1) read(unit=pspline,fmt=*) ii,pshead%lambda,pshead%sigma
   if (pshead%shape_type==3) pshead%shape_type=-1
   pshead%rc_hat=zero
 else if (pspversion>=3) then
!  Shapefunction parameters for Abinit 4.6+
   pshead%rc_hat=zero
   if (pshead%shape_type==-1) read(unit=pspline,fmt=*,err=21,end=21) ii,pshead%rc_hat
   if (pshead%shape_type== 1) read(unit=pspline,fmt=*,err=21,end=21) ii,pshead%rc_hat, &
&   pshead%lambda,pshead%sigma
   if (pshead%shape_type== 2) read(unit=pspline,fmt=*,err=21,end=21) ii,pshead%rc_hat
   if (pshead%shape_type== 3) read(unit=pspline,fmt=*,err=21,end=21) ii,pshead%rc_hat
 end if
 21 continue
 rshp0=pshead%rc_hat 
!If shapefunction type is gaussian, check exponent
 if (pshead%shape_type==1) then
   if (pshead%lambda<2) then
     write(std_out, '(3a)' )&
&     '  For a gaussian shape function, exponent lambda must be >1!',ch10,&
&     '  Action: check your psp file.'
     stop
   end if
 end if
!rc_hat=zero, means rc_hat=r_paw
 rc=pshead%rc_hat;if (rc<1.d-8) pshead%rc_hat=pshead%rc_sph

!------------------------------------------------------------------
!--------------- Partial waves, pseudo partial waves, projectors (phi)

 do ib=1,pshead%basis_size
   read (in_unit,*)
   if (pspversion==1) iread1=1
   if (pspversion>1) read (in_unit,*) iread1
   if (ib==1) then
     imainmesh=iread1
     allocate(pawps%phi(radmesh(imainmesh)%mesh_size,pshead%basis_size))
   else if (iread1/=imainmesh) then
     write(std_out, '(a,a,a)' )&
&     '  All Phi and tPhi must be given on the same radial mesh!',ch10,&
&     '  Action: check your pseudopotential file.'
     stop
   end if
   read (in_unit,*) (pawps%phi(ir,ib),ir=1,radmesh(imainmesh)%mesh_size)
 end do
 mesh_indx(1)=imainmesh

 allocate(pawps%tphi(radmesh(imainmesh)%mesh_size,pshead%basis_size))
 do ib=1,pshead%basis_size
   read (in_unit,*)
   if (pspversion==1) iread1=1
   if (pspversion>1) read (in_unit,*) iread1
   if (iread1/=imainmesh) then
     write(std_out, '(a,a,a)' )&
&     '  All Phi and tPhi must be given on the same radial mesh!',ch10,&
&     '  Action: check your pseudopotential file.'
     stop
   end if
   read (in_unit,*) (pawps%tphi(ir,ib),ir=1,radmesh(imainmesh)%mesh_size)
 end do

 jj=0
 do ib=1,pshead%basis_size
   read (in_unit,*)
   if (pspversion==1) iread1=2
   if (pspversion>1) read (in_unit,*) iread1
   if (ib==1) then
     iprojmesh=iread1
     allocate(pawps%tproj(radmesh(iread1)%mesh_size,pshead%basis_size))
   else if (iread1/=iprojmesh) then
     write(std_out, '(a,a,a)' )&
&     '  All tprojectors must be given on the same radial mesh!',ch10,&
&     '  Action: check your pseudopotential file.'
     stop
   end if
   if(iprojmesh /= -1) then 
    ! read projectors from a mesh
    read (in_unit,*) (pawps%tproj(ir,ib),ir=1,radmesh(iprojmesh)%mesh_size)
   end if
 end do
 mesh_indx(2)=iprojmesh

!------------------------------------------------------------------
!--------------- Core density and pseudo core density

 read (in_unit,*)
 if (pspversion==1) iread1=1
 if (pspversion>1) read (in_unit,*) iread1
 icoremesh=iread1
 if ((radmesh(icoremesh)%mesh_type/=radmesh(imainmesh)%mesh_type).or.&
& (radmesh(icoremesh)%rstep    /=radmesh(imainmesh)%rstep)    .or.&
& (radmesh(icoremesh)%lstep    /=radmesh(imainmesh)%lstep)) then
   write(std_out, '(a,a,a,a,a)' )&
&   '  Ncore must be given on a radial mesh with the same',ch10,&
&   '  type and step(s) than the main radial mesh (mesh for Phi)!',ch10,&
&   '  Action: check your pseudopotential file.'
   stop
 end if
 allocate(pawps%coredens(radmesh(icoremesh)%mesh_size))
 read (in_unit,*) (pawps%coredens(ir),ir=1,radmesh(icoremesh)%mesh_size)
 mesh_indx(3)=icoremesh

 read (in_unit,*)
 if (pspversion==1) iread1=1
 if (pspversion>1) read (in_unit,*) iread1
 if (iread1/=icoremesh) then
   write(std_out, '(a,a,a,a,a)' )&
&   '  Pseudized core density (tNcore) must be given',ch10,&
&   '  on the same radial mesh as core density (Ncore)!',ch10,&
&   '  Action: check your pseudopotential file.'
   stop
 end if
 allocate(pawps%tcoredens(radmesh(icoremesh)%mesh_size))
 read (in_unit,*) (pawps%tcoredens(ir),ir=1,radmesh(icoremesh)%mesh_size)

!------------------------------------------------------------------
!--------------- Frozen part of Dij terms (dij0)

 allocate(pawps%dij0(pshead%lmn2_size))
 read (in_unit,*)
 read (in_unit,*) (pawps%dij0(ib),ib=1,pshead%lmn2_size)

!------------------------------------------------------------------
!--------------- Initial guess of rhoij (rhoij0)

 allocate(pawps%rhoij0(pshead%lmn2_size))
 read (in_unit,*)
 read (in_unit,*) (pawps%rhoij0(ib),ib=1,pshead%lmn2_size)

!------------------------------------------------------------------
!--------------- Local pseudopotential=Vh(tn_zc) or Vbare

 read (in_unit,*)
 if (pspversion==1) ivlocmesh=3
 pshead%vlocopt=1
 if (pspversion==2) then
   read (in_unit,*) ivlocmesh
 else if (pspversion>2) then
   message=blank
   read (in_unit,fmt='(a)') message
   read (message,fmt=*) ivlocmesh
   write(numb,'(i1)')ivlocmesh
   ii=index(message,numb)
   if(len_trim(trim(message(ii+1:)))/=0)then
     submessage=trim(message(ii+1:))
     if(len_trim(submessage)/=0)then
       do ii=1,len_trim(submessage)
         numb=submessage(ii:ii)
         if(numb==blank)cycle
         jj=index('0123456789',numb)
         if(jj<1 .or. jj>10)exit
         pshead%vlocopt=jj-1
       end do
     end if
   end if
 end if
 allocate(pawps%vhtnzc(radmesh(ivlocmesh)%mesh_size))
 read (in_unit,*) (pawps%vhtnzc(ir),ir=1,radmesh(ivlocmesh)%mesh_size)
 mesh_indx(4)=ivlocmesh

!------------------------------------------------------------------
!--------------- "numeric" shapefunctions (if shape_type=-1)

 if (pshead%shape_type==-1) then
   do il=1,pshead%l_size
     read (in_unit,*)
     if (pspversion==1) iread1=1
     if (pspversion>1) read (in_unit,*) iread1
     if (il==1) then
       ishpfmesh=iread1
       allocate(pawps%shpf(radmesh(ishpfmesh)%mesh_size,pshead%l_size))
     else if (iread1/=ishpfmesh) then
       write(std_out, '(a,a,a)' )&
&       '  All shape functions must be given on the same radial mesh!',ch10,&
&       '  Action: check your pseudopotential file.'
       stop
     end if
     read (in_unit,*) (pawps%shpf(ir,il),ir=1,radmesh(ishpfmesh)%mesh_size)
   end do
 end if
 mesh_indx(5)=ishpfmesh

!------------------------------------------------------------------
!--------------- Pseudo valence density (if psp version >=4)

 if (pspversion>=4) then
   read (in_unit,*)
   read (in_unit,*) iread1
   ivalemesh=iread1
   if (radmesh(ivalemesh)%rad(radmesh(ivalemesh)%mesh_size) &
&     <radmesh(imainmesh)%rad(radmesh(imainmesh)%mesh_size)) then
     write(std_out, '(a,a,a,a,a)' )&
&     '  Pseudized valence density (tNvale) must be given',ch10,&
&     '  on the radial mesh greater than the one for the basis functions (Phi and tPhi)!',ch10,&
&     '  Action: check your pseudopotential file.'
     stop
   end if
   allocate(pawps%tvaledens(radmesh(ivalemesh)%mesh_size))
   read (in_unit,*) (pawps%tvaledens(ir),ir=1,radmesh(ivalemesh)%mesh_size)
   usetvale=1
 else
   usetvale=0
 end if
 mesh_indx(6)=ivalemesh

!------------------------------------------------------------------

 close(in_unit)

 end subroutine rdpawps


!================================================================================
! ROUTINE: WRITING OF THE XML-FORMATTED FILE
!================================================================================

!! FUNCTION
!!  Write the paw data file in XML format for ABINIT
!!
!! INPUTS
!!  xml_file_towrite=name of the output XML-formatted PAW dataset
!!  mesh_indx(:)= indexes the different meshes
!!  pshead
!!    %atomic_charge= Total atomic charge
!!    %basis_size= Number of elements for the paw nl basis
!!    %core_charge= Core charge
!!    %vlocopt=Option for the format of Vloc
!!              0: Vloc is Vbare (Blochl formulation)
!!              1: Vloc is VH(tnzc) (Kresse formulation)
!!              2: Vloc is VH(tnzc) (Kresse formulation without compensation in XC)
!!    %lambda= Lambda in gaussian type g(r)
!!    %lmn_size= Number of elements for the paw basis
!!    %lmn2_size=lmn_size*(lmn_size+1)/2
!!    %l_size = Maximum value of l+1 leading to a non zero Gaunt coef
!!    %max_meshsz   = max. number of radial meshes
!!    %orbitals(basis_size)= Quantum number l for each basis function
!!    %pspxc_abinit= ABINIT code number for the exchange-correlation
!!    %pspversion= Version of the PAW format for Abinit
!!    %rc_hat= radius for shape function
!!    %rc_sph= Default PAW sphere radius
!!    %shape_type= Shape function type
!!    %sigma= Sigma for gaussian type g(r)
!!    %title= Title for pseudopotential
!! pawps
!!  %coredens(:)  ! Core density
!!  %tcoredens(:) ! Pseudo core density
!!  %tvaledens(:) ! Pseudo valence density multiplied
!!  %phi(:,:)     ! PAW atomic partial waves
!!  %tphi(:,:)    ! PAW atomic pseudo partial waves
!!  %tproj(:,:)   ! PAW projectors for each basis function
!!  %dij0(:)      ! Part of the Dij term frozen during SCF
!!  %kij(:)       ! Kinetic overlap operator
!!  %rhoij0(:)    ! Atomic initialization of rhoij
!!  %shpf(:,:)    ! l-dependent shapefunction
!!  %vhtnzc(:)    ! Hartree potential of the ps-density of (nucleus+core_el)

 subroutine wrxmlps(xml_file_towrite,mesh_indx,pawps,pshead,radmesh)

!Arguments
 integer,intent(out) :: mesh_indx(nmesh_max)
 character(len=fnlen),intent(in) :: xml_file_towrite
 type(pshead_type),intent(inout) :: pshead
 type(pawrad_type),intent(inout) :: radmesh(nmesh_max)
 type(pawps_type),intent(inout)  :: pawps

!Local variables
 integer, parameter :: out_unit=11
 integer :: ii,ib,il,iln,ilm,ilmn,ilmn0,ir,lmnmax,nbase,OK
 integer :: jlmn,jlm,jln,j0lmn,klmn,nmesh,zz
 integer :: imainmesh,iprojmesh,ishpfmesh,icoremesh,iprjmesh,ivalemesh,ivlocmesh
 real(dp) :: radstp0,logstp0,sqr4pi,rc_hat
 character (len=4) :: flnm,flnm1
 character (len=20) :: nm,nm1
 character(len=5) ::f_type
 character(len=12) ::f_name
 integer :: mesh_start(nmesh_max)
 integer,allocatable :: indlmn(:,:),nprj(:)
 real(dp), allocatable :: dum(:),ff(:),kij(:),kij0(:,:)
 character(len=3) :: gridt(nmesh_max)

!------------------------------------------------------------------

!------------------------------------------------------------------
!--------------- Useful constants

 nmesh=pshead%nmesh
 sqr4pi=sqrt(4*pi)
 nbase=pshead%basis_size
 zz=nint(pshead%atomic_charge)
 call find_atomname (zz,flnm)

 open(out_unit,file=trim(xml_file_towrite),form='formatted')

!------------------------------------------------------------------
!--------------- Header

 write(out_unit,'("<?xml  version=""1.0""?>")')
 write(out_unit,'("<paw_dataset version=""0.7"">")')
 WRITE(out_unit,'("<!-- PAW-XML specification: http://esl.cecam.org/Paw-xml -->")')

 write(out_unit,'("<atom symbol=""",a,""" Z=""",f5.2,$)') trim(flnm),pshead%atomic_charge
 write(out_unit,'(""" core=""", f5.2,""" valence=""",f5.2"""/>")')pshead%core_charge,&
&                                            (pshead%atomic_charge-pshead%core_charge)

!------------------------------------------------------------------
!--------------- XC functional

 if(pshead%pspxc_abinit<0) then
   f_type="LIBXC"
   write(f_name,'(i6)') -pshead%pspxc_abinit
 end if

 select case(pshead%pspxc_abinit)
   case(2,-9001,-1009)
     f_type="LDA"
     f_name="PZ"
   case(4,-2001,-1002)
     f_type="LDA"
     f_name="W"
   case(5,-4001,-1004)
     f_type="LDA"
     f_name="HL"
   case(7,-12001,-1012)
     f_type="LDA"
     f_name="PW"
   case(11,-101130,-130101)
     f_type="GGA"
     f_name="PBE"
   case(14,-102130,-130102)
     f_type="GGA"
     f_name="revPBE"
   case(15,-117130,-130117)
     f_type="GGA"
     f_name="RPBE"
   case(-101116,-116101)
     f_type="GGA"
     f_name="PBEsol"
   case(-134109,-109134)
     f_type="GGA"
     f_name="PW91"
   case(-135120,-120135)
     f_type="GGA"
     f_name="AM05"
   case(-131106,-106131)
     f_type="GGA"
     f_name="BLYP"
   case default
     write(std_out, '(a,a,a)' )&
&       '  Unable to identify XC functional in input file!',ch10,&
&       '  Action: check your input pseudopotential file.'
     stop
 end select

 write(out_unit,'("<xc_functional type=""",a,""" name=""",a,"""/>")')TRIM(f_type),TRIM(f_name)
 write(out_unit,'("<generator type=""translator"" name=""abinit2xml"">")')
 write(out_unit,'("</generator>")')

!------------------------------------------------------------------
!--------------- Title, PAW radius

 write(out_unit,'("<!--")')
 write(out_unit,'("  ",a)') trim(pshead%title)
 write(out_unit,'("  PAW functions generated on ",a,"/",a,"/",a)') trim(pshead%pspdat(5:6)),trim(pshead%pspdat(7:8)),&
&                 trim(pshead%pspdat(1:4))
 write(out_unit,'("  -->")')
 write(out_unit,'("<paw_radius rc=""",f17.14,"""/>")') pshead%rc_sph


!------------------------------------------------------------------
!--------------- Valence states

 write(out_unit,'("<valence_states>")')

 allocate(ff(nbase))
 ff=zero
 ilmn0=0
 do ib=1,nbase
   il=2*pshead%orbitals(ib)+1
   do ilmn=ilmn0+1,ilmn0+il
     ff(ib)=ff(ib)+pawps%rhoij0(ilmn*(ilmn+1)/2)
   end do
   ilmn0=ilmn0+il
 end do

 do ib=1,nbase
    call mkname(ib,flnm1)
    nm=stripchar('"'//flnm//flnm1//'"')
    write(out_unit,'("  <state        l=""",i1,""" f=""",1pe14.7,$)')&
&        pshead%orbitals(ib),ff(ib)
    write(out_unit,'(""" rc=""",f13.10,""" e=""",1pe14.7,""" id=",a6,"/>")')&
&        pshead%rc_sph,0.d0,TRIM(nm)
 enddo
 deallocate (ff)
 deallocate(pawps%rhoij0)

 write(out_unit,'("</valence_states>")')


!------------------------------------------------------------------
!--------------- Radial meshes

 do ii=1,nmesh
   allocate(dum(radmesh(ii)%mesh_size))
   select case(radmesh(ii)%mesh_type)
     case(1)
       nm='r=d*i'
       gridt(ii)="lin"
       mesh_start(ii)=1
       radstp0=zero
       logstp0=radmesh(ii)%rstep
       dum=logstp0
     case(2)
       nm='r=a*(exp(d*i)-1)'
       gridt(ii)="log"
       mesh_start(ii)=1
       radstp0=radmesh(ii)%rstep
       logstp0=radmesh(ii)%lstep
       dum(:)=logstp0*(radstp0+radmesh(ii)%rad(:)) 
     case(3)
       nm='r=a*exp(d*i)'
       gridt(ii)="log"
       mesh_start(ii)=1;if (radmesh(ii)%rad(1)<tol12) mesh_start(ii)=2
       radstp0=radmesh(ii)%rstep
       logstp0=radmesh(ii)%lstep
       dum(:)=logstp0*radmesh(ii)%rad(:) 
     case default
       write(std_out, '(a,a,a)' )&
&       '  This mesh type ', ch10,&
&       '  is not implemented in Atompaw.'
        stop
   end select
   write(out_unit,'("<radial_grid eq=""",a,""" a=""",e23.16,$)') trim(nm),radstp0
   write(out_unit,'(""" d=""",e23.16,""" istart=""0"" iend=""",i5,$)') logstp0,radmesh(ii)%mesh_size-mesh_start(ii)
   write(out_unit,'(""" id=""",a,i1,""">")') gridt(ii),ii
   write(out_unit,'("  <values>")')
   write(out_unit,'(3(1x,es23.16))') (radmesh(ii)%rad(ir),ir=mesh_start(ii),radmesh(ii)%mesh_size)
   write(out_unit,'("  </values>")')
   write(out_unit,'("  <derivatives>")')
   write(out_unit,'(3(1x,es23.16))') (dum(ir),ir=mesh_start(ii),radmesh(ii)%mesh_size)
   write(out_unit,'("  </derivatives>")')
   write(out_unit,'("</radial_grid>")')
   deallocate(dum)
 end do


!------------------------------------------------------------------
!--------------- Compensation charge

 rc_hat=pshead%rc_hat
 select case(pshead%shape_type)
   case(1)
     nm='exp'
     if(pshead%lambda==2) nm='gauss'
     if(nm=='exp') rc_hat=pshead%sigma
   case(2)
     nm='sinc'
   case(3)
     nm='bessel'
   case(-1)
     nm='num'
 end select

 if((pshead%shape_type==1).and.(pshead%lambda/=2)) then
   write(out_unit,'("<shape_function type=""",a,""" rc=""",f19.16,""" lamb=""",i3,"""/>")')&
 &          trim(nm),rc_hat,pshead%lambda
 end if
 if((pshead%shape_type==2).or.(pshead%shape_type==3).or.(pshead%lambda==2)) then
   write(out_unit,'("<shape_function type=""",a,""" rc=""",f19.16"""/>")') trim(nm),rc_hat
 end if
 if(pshead%shape_type==-1) then
   ishpfmesh=mesh_indx(5)
   call mkname(ishpfmesh,flnm1)
   nm1=stripchar('"'//gridt(ishpfmesh)//flnm1//'"')
   do il=1,pshead%l_size
     write(out_unit,'("<shape_function type=""",a,""" grid=",a," l=""",i1,""">")')&
 &          trim(nm),trim(nm1),il-1
     write(out_unit,*) (pawps%shpf(ir,il),ir=mesh_start(ishpfmesh),radmesh(ishpfmesh)%mesh_size)
     write(out_unit,'("</shape_function>")')
   end do
 end if


!------------------------------------------------------------------
!--------------- Core density and pseudo core density

 icoremesh=mesh_indx(3)
 write(out_unit,'("<ae_core_density grid=""",a,i1""">")') gridt(icoremesh),icoremesh
 write(out_unit,'(3(1x,es23.16))') (sqr4pi*pawps%coredens(ir),ir=mesh_start(icoremesh),radmesh(icoremesh)%mesh_size)
 write(out_unit,'("</ae_core_density>")')

 write(out_unit,'("<pseudo_core_density grid=""",a,i1,""">")') gridt(icoremesh),icoremesh
 write(out_unit,'(3(1x,es23.16))') (sqr4pi*pawps%tcoredens(ir),ir=mesh_start(icoremesh),radmesh(icoremesh)%mesh_size)
 write(out_unit,'("</pseudo_core_density>")')


!------------------------------------------------------------------
!--------------- Valence density

 if (pshead%pspversion>=4) then
   ivalemesh=mesh_indx(6)
   write(out_unit,'("<pseudo_valence_density grid=""",a,i1,""" rc=""",f19.16,""">")') gridt(ivalemesh),ivalemesh,pshead%rc_sph
   write(out_unit,'(3(1x,es23.16))') (sqr4pi*pawps%tvaledens(ir),ir=mesh_start(ivalemesh),radmesh(ivalemesh)%mesh_size)
   write(out_unit,'("</pseudo_valence_density>")')
 end if


!------------------------------------------------------------------
!--------------- Local potential(s)

 ivlocmesh=mesh_indx(4)
 if(pshead%vlocopt==1) then
   write(out_unit,'("<kresse_joubert_local_ionic_potential grid=""",a,i1,""">")')&
&       gridt(ivlocmesh),ivlocmesh
   write(out_unit,'(3(1x,es23.16))') (sqr4pi*pawps%vhtnzc(ir),ir=mesh_start(ivlocmesh),radmesh(ivlocmesh)%mesh_size)
   write(out_unit,'("</kresse_joubert_local_ionic_potential>")')
 end if
 if(pshead%vlocopt==2) then
   write(out_unit,'("<blochl_local_ionic_potential grid=""",a,i1,""">")')&
&       gridt(ivlocmesh),ivlocmesh
   write(out_unit,'(3(1x,es23.16))') (sqr4pi*pawps%vhtnzc(ir),ir=mesh_start(ivlocmesh),radmesh(ivlocmesh)%mesh_size)
   write(out_unit,'("</blochl_local_ionic_potential>")')
 endif
  

!------------------------------------------------------------------
!--------------- Partial waves and Projectors

 imainmesh=mesh_indx(1)
 iprojmesh=mesh_indx(2)
 do ib=1,nbase
   allocate(dum(radmesh(imainmesh)%mesh_size),stat=OK)
   call mkname(ib,flnm1)
   nm=stripchar('"'//flnm//flnm1//'"')
   dum=0
   dum(2:radmesh(imainmesh)%mesh_size)=pawps%phi(2:radmesh(imainmesh)%mesh_size,ib)&
&           /radmesh(imainmesh)%rad(2:radmesh(imainmesh)%mesh_size)
   call deducer0(dum,radmesh(imainmesh)%mesh_size,radmesh(imainmesh))
   write(out_unit,'("<ae_partial_wave state=",a6," grid=""",a,i1,""">")')&
&          trim(nm),gridt(imainmesh),imainmesh
   write(out_unit,'(3(1x,es23.16))') (dum(ii),ii=mesh_start(imainmesh),radmesh(imainmesh)%mesh_size)
   write(out_unit,'("</ae_partial_wave>")')
   dum=0
   dum(2:radmesh(imainmesh)%mesh_size)=pawps%tphi(2:radmesh(imainmesh)%mesh_size,ib)&
&           /radmesh(imainmesh)%rad(2:radmesh(imainmesh)%mesh_size)
   call deducer0(dum,radmesh(imainmesh)%mesh_size,radmesh(imainmesh))
   write(out_unit,'("<pseudo_partial_wave state=",a6," grid=""",a,i1,""">")')&
&          trim(nm),gridt(imainmesh),imainmesh
   write(out_unit,'(3(1x,es23.16))') (dum(ii),ii=mesh_start(imainmesh),radmesh(imainmesh)%mesh_size)
   write(out_unit,'("</pseudo_partial_wave>")')
   deallocate(dum)
   allocate(dum(radmesh(iprojmesh)%mesh_size),stat=OK)
   dum=0
   dum(2:radmesh(iprojmesh)%mesh_size)=pawps%tproj(2:radmesh(iprojmesh)%mesh_size,ib)&
&           /radmesh(iprojmesh)%rad(2:radmesh(iprojmesh)%mesh_size)
   call deducer0(dum,radmesh(iprojmesh)%mesh_size,radmesh(iprojmesh))
   write(out_unit,'("<projector_function state=",a6," grid=""",a,i1,""">")') &
&            trim(nm),gridt(iprojmesh),iprojmesh
   write(out_unit,'(3(1x,es23.16))') (dum(ii),ii=mesh_start(iprojmesh),radmesh(iprojmesh)%mesh_size)
   write(out_unit,'("</projector_function>")')
   deallocate (dum)
 end do

!------------------------------------------------------------------
!--------------- Kinetic energy differences

 allocate (kij(pshead%lmn2_size))
 allocate(nprj(0:maxval(pshead%orbitals)))
 lmnmax=0
 do il=1,pshead%basis_size
   lmnmax=lmnmax+2*pshead%orbitals(il)+1
 end do
 allocate(indlmn(6,lmnmax))
 ilmn=0;iln=0;nprj=0
 do ib=1,pshead%basis_size
   il=pshead%orbitals(ib)
   nprj(il)=nprj(il)+1
   iln=iln+1
   do ilm=1,2*il+1
     indlmn(1,ilmn+ilm)=il
     indlmn(2,ilmn+ilm)=ilm-(il+1)
     indlmn(3,ilmn+ilm)=nprj(il)
     indlmn(4,ilmn+ilm)=il*il+ilm
     indlmn(5,ilmn+ilm)=iln
     indlmn(6,ilmn+ilm)=1
   end do
   ilmn=ilmn+2*il+1
 end do
 DEALLOCATE(nprj)

 if (allocated(pawps%kij)) then
   kij=pawps%kij
 else
   call pawkij(indlmn,kij,lmnmax,pawps%coredens,1,pawps,pshead,&
&               radmesh(imainmesh),radmesh(icoremesh),radmesh(ivlocmesh),pawps%vhtnzc,pshead%atomic_charge)
 end if

 allocate (kij0(pshead%basis_size,pshead%basis_size))
 kij0=zero
 do jlmn=1,pshead%lmn_size
   j0lmn=jlmn*(jlmn-1)/2
   jlm=indlmn(4,jlmn);jln=indlmn(5,jlmn)
   do ilmn=1,jlmn
     klmn=j0lmn+ilmn
     ilm=indlmn(4,ilmn);iln=indlmn(5,ilmn)
     if (ilm==jlm) then
       kij0(iln,jln)=kij(klmn)
       kij0(jln,iln)=kij(klmn)
     end if
   end do
 end do

 write(out_unit,'("<kinetic_energy_differences>")')
 write(out_unit,'(3(1x,es23.16))') ((kij0(ir,ii),ii=1,pshead%basis_size),ir=1,pshead%basis_size)
 write(out_unit,'("</kinetic_energy_differences>")')
 write(out_unit,'("</paw_dataset>")')

!------------------------------------------------------------------
!--------------- End: memory release

 deallocate(pshead%orbitals)
 deallocate (indlmn)
 deallocate(pawps%phi,pawps%tphi,pawps%tproj)
 deallocate(pawps%coredens,pawps%tcoredens)
 deallocate(pawps%vhtnzc)
 deallocate (kij,kij0)
 deallocate (pawps%dij0)
 if (pshead%shape_type==-1) deallocate(pawps%shpf)
 if (pshead%pspversion>=4) deallocate(pawps%tvaledens)
 do ii=1,nmesh
   call destroy_radmesh(radmesh(ii))
 end do

!------------------------------------------------------------------

 close(out_unit)

 end subroutine wrxmlps


!================================================================================
! ROUTINE: FIND ATOM NAME FROM ATOMIC NUMBER
!================================================================================

!! FUNCTION
!!  Give specie name from atomic number
!!

 subroutine find_atomname(zz,flnm)

!Arguments
 integer,intent(in) :: zz
 character (len=4), intent(out) :: flnm

!------------------------------------------------------------------

 select case(zz)
   case(1)
     flnm='H'
   case(2)
     flnm='He'
   case(3)
     flnm='Li'
   case(4)
     flnm='Be'
   case(5)
     flnm='B'
   case(6)
     flnm='C'
   case(7)
     flnm='N'
   case(8)
     flnm='O'
   case(9)
     flnm='F'
   case(10)
     flnm='Ne'
   case(11)
     flnm='Na'
   case(12)
     flnm='Mg'
   case(13)
     flnm='Al'
   case(14)
     flnm='Si'
   case(15)
     flnm='P'
   case(16)
     flnm='S'
   case(17)
     flnm='Cl'
   case(18)
     flnm='Ar'
   case(19)
     flnm='K'
   case(20)
     flnm='Ca'
   case(21)
     flnm='Sc'
   case(22)
     flnm='Ti'
   case(23)
     flnm='V'
   case(24)
     flnm='Cr'
   case(25)
     flnm='Mn'
   case(26)
     flnm='Fe'
   case(27)
     flnm='Co'
   case(28)
     flnm='Ni'
   case(29)
     flnm='Cu'
   case(30)
     flnm='Zn'
   case(31)
     flnm='Ga'
   case(32)
     flnm='Ge'
   case(33)
     flnm='As'
   case(34)
     flnm='Se'
   case(35)
     flnm='Br'
   case(36)
     flnm='Kr'
   case(37)
     flnm='Rb'
   case(38)
     flnm='Sr'
   case(39)
     flnm='Y'
   case(40)
     flnm='Zr'
   case(41)
     flnm='Nb'
   case(42)
     flnm='Mo'
   case(43)
     flnm='Tc'
   case(44)
     flnm='Ru'
   case(45)
     flnm='Rh'
   case(46)
     flnm='Pd'
   case(47)
     flnm='Ag'
   case(48)
     flnm='Cd'
   case(49)
     flnm='In'
   case(50)
     flnm='Sn'
   case(51)
     flnm='Sb'
   case(52)
     flnm='Te'
   case(53)
     flnm='I'
   case(54)
     flnm='Xe'
   case(55)
     flnm='Cs'
   case(56)
     flnm='Ba'
   case(57)
     flnm='La'
   case(58)
     flnm='Ce'
   case(59)
     flnm='Pr'
   case(60)
     flnm='Nd'
   case(61)
     flnm='Pm'
   case(62)
     flnm='Sm'
   case(63)
     flnm='Eu'
   case(64)
     flnm='Gd'
   case(65)
     flnm='Tb'
   case(66)
     flnm='Dy'
   case(67)
     flnm='Ho'
   case(68)
     flnm='Er'
   case(69)
     flnm='Tm'
   case(70)
     flnm='Yb'
   case(71)
     flnm='Lu'
   case(72)
     flnm='Hf'
   case(73)
     flnm='Ta'
   case(74)
     flnm='W'
   case(75)
     flnm='Re'
   case(76)
     flnm='Os'
   case(77)
     flnm='Ir'
   case(78)
     flnm='Pt'
   case(79)
     flnm='Au'
   case(80)
     flnm='Hg'
   case(81)
     flnm='Tl'
   case(82)
     flnm='Pb'
   case(83)
     flnm='Bi'
   case(84)
     flnm='Po'
   case(85)
     flnm='At'
   case(86)
     flnm='Rn'
   case(87)
     flnm='Fr'
   case(88)
     flnm='Ra'
   case(89)
     flnm='Ac'
   case(90)
     flnm='Th'
   case(91)
     flnm='Pa'
   case(92)
     flnm='U'
   case(93)
     flnm='Np'
   case(94)
     flnm='Pu'
   case(95)
     flnm='Am'
   case(96)
     flnm='Cm'
   case(97)
     flnm='Bk'
   case(98)
     flnm='Cf'
   case(99)
     flnm='Es'
   case(100)
     flnm='Fm'
   case(101)
     flnm='Md'
   case(102)
     flnm='No'
   case(103)
     flnm='Lw'
 end select

 end subroutine find_atomname


!================================================================================
! ROUTINE: COMPUTE KINETIC PART OF DIJ
!================================================================================

!! FUNCTION
!! Deduce kinetic part of psp strength (Dij) from the knowledge of frozen Dij (Dij0)
!!
!!
!! INPUTS
!!  indlmn(6,lmnmax)= array giving l,m,n,lm,ln,s for i=lmn
!!  lmnmax=max number of (l,m,n) components over all type of psps
!!  ncore(radmesh_core%mesh_size)=atomic core density
!!  opt_init=flag defining the storage of PAW atomic data
!!           0: PAW atomic data have not been initialized (in pawtab)
!!           1: PAW atomic data have been initialized (in pawtab)
!!  opt_vhnzc=flag defining the inclusion of VH(nZc) in computation
!!            0: VH(nZc) is not taken into account
!!            1: VH(nZc) is taken into account
!!  pawtab <type(pawtab_type)>=paw tabulated starting data
!!  radmesh <type(pawrad_type)>=paw radial mesh (and related data)
!!  radmesh_core <type(pawrad_type)>=radial mesh (and related data) for the core densities
!!  radmesh_vloc <type(pawrad_type)>=radial mesh (and related data) for the local potential (VH(tnZc))
!!  vhtnzc(radmesh_vloc%mesh_size)= local potential VH(tnZc)
!!  znucl= valence and total charge of the atomic species
!!
!! OUTPUT
!!  kij(pawtab%lmn2_size)= kinetic part of Dij

 subroutine pawkij(indlmn,kij,lmnmax,ncore,opt_vhnzc,pawps,pshead,radmesh,radmesh_core,radmesh_vloc,vhtnzc,znucl)

!Arguments
 integer,intent(in) :: lmnmax,opt_vhnzc
 real(dp),intent(in) :: znucl
 type(pawrad_type),intent(in) :: radmesh,radmesh_core,radmesh_vloc
 type(pawps_type),intent(in) :: pawps
 type(pshead_type),intent(in) :: pshead
 integer,intent(in) :: indlmn(6,lmnmax)
 real(dp),intent(out) :: kij(pshead%lmn2_size)
 real(dp),intent(in) :: ncore(radmesh_core%mesh_size)
 real(dp),intent(in) :: vhtnzc(radmesh_vloc%mesh_size)

!Local variables
 integer :: il,ilm,iln,ilmn,j0lmn,jl,jlm,jln,jlmn,klmn,lmn2_size,meshsz
 real(dp) :: intg,intvh,yp1,ypn
 real(dp),allocatable :: ff(:),shpf(:),vhnzc(:),vhtnzc_sph(:),r2k(:),work1(:),work2(:)

!------------------------------------------------------------------

 lmn2_size=pshead%lmn2_size
 meshsz=min(radmesh%mesh_size,radmesh_core%mesh_size,radmesh_vloc%mesh_size,ifromr(radmesh,pshead%rc_sph))
 ALLOCATE(ff(meshsz))

!Retrieve VH(tnZc) on the correct radial mesh
 ALLOCATE(vhtnzc_sph(meshsz))
 if ((radmesh%mesh_type/=radmesh_vloc%mesh_type).or.&
&    (radmesh%rstep    /=radmesh_vloc%rstep)    .or.&
&    (radmesh%lstep    /=radmesh_vloc%lstep)) then
   call bound_deriv(vhtnzc(1:radmesh_vloc%mesh_size),radmesh_vloc,radmesh_vloc%mesh_size,yp1,ypn)
   ALLOCATE(work1(radmesh_vloc%mesh_size))
   ALLOCATE(work2(radmesh_vloc%mesh_size))
   call spline(radmesh_vloc%rad,vhtnzc,radmesh_vloc%mesh_size,yp1,ypn,work1)
   call splint(radmesh_vloc%mesh_size,radmesh_vloc%rad,vhtnzc,work1,meshsz,radmesh%rad(1:meshsz),vhtnzc_sph)
   DEALLOCATE(work1)
   DEALLOCATE(work2)
 else
   vhtnzc_sph(1:meshsz)=vhtnzc(1:meshsz)
 end if

!Initiialize Kij with Dij0
 kij(1:lmn2_size)=pawps%dij0(1:lmn2_size)

!Substraction of <phi_i|vh(nZc)|phi_j> on the PAW sphere
 if (opt_vhnzc/=0) then
   ALLOCATE(vhnzc(radmesh_core%mesh_size))
   call pawvhnzc(ncore,radmesh_core,vhnzc,znucl)
   do jlmn=1,pshead%lmn_size
     j0lmn=jlmn*(jlmn-1)/2
     jlm=indlmn(4,jlmn);jln=indlmn(5,jlmn)
     do ilmn=1,jlmn
       klmn=j0lmn+ilmn
       ilm=indlmn(4,ilmn);iln=indlmn(5,ilmn)
       if (jlm==ilm) then
         ff(1:meshsz)=pawps%phi(1:meshsz,iln)*pawps%phi(1:meshsz,jln)*vhnzc(1:meshsz)
         call simp_gen(intg,ff,radmesh,r_for_intg=pshead%rc_sph)
         kij(klmn)=kij(klmn)-intg
       end if
     end do
   end do
   DEALLOCATE(vhnzc)
 end if

!Substraction of -<tphi_i|vh(tnZc)|tphi_j> on the PAW sphere
 do jlmn=1,pshead%lmn_size
   j0lmn=jlmn*(jlmn-1)/2
   jlm=indlmn(4,jlmn);jln=indlmn(5,jlmn)
   do ilmn=1,jlmn
     klmn=j0lmn+ilmn
     ilm=indlmn(4,ilmn);iln=indlmn(5,ilmn)
     if (jlm==ilm) then
       ff(1:meshsz)=pawps%tphi(1:meshsz,iln)*pawps%tphi(1:meshsz,jln)*vhtnzc_sph(1:meshsz)
       call simp_gen(intg,ff,radmesh,r_for_intg=pshead%rc_sph)
       kij(klmn)=kij(klmn)+intg
     end if
   end do
 end do

!Substraction of -int[vh(tnzc)*Qijhat(r)dr]
 ALLOCATE(shpf(radmesh%mesh_size))
 call pawshpfun(0,radmesh,intg,pawps,pshead,shpf)
 if (pshead%shape_type==3) then
   ALLOCATE(r2k(radmesh%int_meshsz))
   r2k=zero
   r2k(2:radmesh%int_meshsz)=shpf(2:radmesh%int_meshsz)*radmesh%rad(2:radmesh%int_meshsz)**2
   if(radmesh%mesh_type==5) then
     call simp_gen(intg,r2k,radmesh)
   else
     call simp_gen(intg,r2k,radmesh,r_for_intg=pshead%rc_hat)
   end if
   shpf(1:meshsz)=shpf(1:meshsz)/intg
   DEALLOCATE(r2k)
 end if
 ff(1:meshsz)=vhtnzc_sph(1:meshsz)*shpf(1:meshsz)*radmesh%rad(1:meshsz)**2
 DEALLOCATE(shpf)
 call simp_gen(intvh,ff,radmesh,r_for_intg=pshead%rc_sph)
 do jlmn=1,pshead%lmn_size
   j0lmn=jlmn*(jlmn-1)/2
   jl=indlmn(1,jlmn);jln=indlmn(5,jlmn);jlm=indlmn(4,jlmn)
   do ilmn=1,jlmn
     klmn=j0lmn+ilmn
     il=indlmn(1,ilmn);iln=indlmn(5,ilmn);ilm=indlmn(4,ilmn)
     if (ilm==jlm) then
       ff(1:meshsz)=(pawps%phi (1:meshsz,iln)*pawps%phi (1:meshsz,jln)&
&                   -pawps%tphi(1:meshsz,iln)*pawps%tphi(1:meshsz,jln))
       call simp_gen(intg,ff,radmesh,r_for_intg=pshead%rc_sph)
       kij(klmn)=kij(klmn)+intvh*intg
     end if
   end do
 end do

 DEALLOCATE(ff)
 DEALLOCATE(vhtnzc_sph)

 end subroutine pawkij


!================================================================================
! ROUTINE: COMPUTE IONIC POTENTIAL VH(tNZC)
!================================================================================

!! FUNCTION
!! PAW: compute Hartree potential for n_{Zc}
!!
!! INPUTS
!!  ncore(radmesh_core%mesh_size)=atomic core density
!!  radmesh_core <type(pawrad_type)>=radial mesh (and related data) for the core densities
!!  znucl= valence and total charge of the atomic species
!!
!! OUTPUT
!!  vhnzc(radmesh_core%mesh_size)=Hartree potential due to Z_nc

 subroutine pawvhnzc(ncore,radmesh_core,vhnzc,znucl)

!Arguments
 real(dp),intent(in) :: znucl
 type(pawrad_type),intent(in) :: radmesh_core
 real(dp),intent(in) :: ncore(radmesh_core%mesh_size)
 real(dp), intent(out) :: vhnzc(radmesh_core%mesh_size)

!Local variables
  real(dp) :: intg
  real(dp),allocatable :: nwk(:)

!------------------------------------------------------------------

 ALLOCATE(nwk(radmesh_core%mesh_size))
 nwk(:)=ncore(:)*four_pi*radmesh_core%rad(:)**2
 call poisson(nwk,0,intg,radmesh_core,vhnzc)
 vhnzc(2:radmesh_core%mesh_size)=(vhnzc(2:radmesh_core%mesh_size)-znucl)/radmesh_core%rad(2:radmesh_core%mesh_size)
 call deducer0(vhnzc,radmesh_core%mesh_size,radmesh_core)
 DEALLOCATE(nwk)

 end subroutine pawvhnzc


!================================================================================
! ROUTINE: COMPUTE SHAPE FUNCTION FOR COMPENSATION CHARGE
!================================================================================

!! FUNCTION
!! Compute shape function used in the definition
!! of compensation density (PAW)
!!
!! INPUTS
!!  ll= l quantum number
!!  mesh <type(pawrad_type)>=data containing radial grid information
!!  pawtab <type(pawtab_type)>=paw tabulated starting data
!!
!! OUTPUT
!!  norm= factor for shape function normalization
!!
!! SIDE effects
!!  shapefunc(mesh%mesh_size)=shape function g(r)
!!    In case of numerical shape function (shape_type=-1), shapefunc
!!    array contains the shape function read in psp file at input.
!!
!! NOTES
!!  Types of shape functions:
!!   type -1: numerical shape function, given in psp file
!!   type  1: g(r)=k(r).r^l; k(r)=exp(-(r/sigma)^lambda)
!!   type  2: g(r)=k(r).r^l; k(r)=[sin(Pi.r/rshp)/(Pi.r/rshp)]^2
!!   type  3: g(r)=alpha1.jl(q1.r)+alpha2.jl(q2.r)

 subroutine pawshpfun(ll,mesh,norm,pawps,pshead,shapefunc)

!Arguments
 integer,intent(in) :: ll
 real(dp),intent(out) :: norm
 type(pawrad_type),intent(in) :: mesh
 type(pawps_type),intent(in) :: pawps
 type(pshead_type),intent(in) :: pshead
 real(dp),intent(inout) :: shapefunc(mesh%mesh_size)

!Local variables
 integer :: ir,ishp
 real(dp) :: arg,argj1,argj2,besp,bespp,jbes1,jbes2,shapefunc1,shapefunc2
 real(dp) :: shapefunc3
 real(dp) :: alpha(2),qq(2)
 real(dp),allocatable :: r2k(:)

!Statement functions
 shapefunc1(arg)= exp(-(arg/pshead%sigma)**pshead%lambda)
 shapefunc2(arg)= (sin(pi*arg/pshead%rc_hat)/(pi*arg/pshead%rc_hat))**2
 shapefunc3(argj1,argj2)= alpha(1)*argj1+alpha(2)*argj2

!------------------------------------------------------------------

!Index for shape function cut-off radius
 ishp=ifromr(mesh,pshead%rc_hat)-1

!Computation of non-normalized shape function
 if (pshead%shape_type==-1) then
   shapefunc(1:ishp)=pawps%shpf(1:ishp,1+ll)
 else if (pshead%shape_type==1) then
   if (ll==0) then
     shapefunc(1)=one
     do ir=2,ishp
       shapefunc(ir)=shapefunc1(mesh%rad(ir))
     end do
   else
     shapefunc(1)=zero
     do ir=2,ishp
       shapefunc(ir)=shapefunc1(mesh%rad(ir))*mesh%rad(ir)**ll
     end do
   end if
 else if (pshead%shape_type==2) then
   if (ll==0) then
     shapefunc(1)=one
     do ir=2,ishp
       shapefunc(ir)=shapefunc2(mesh%rad(ir))
     end do
   else
     shapefunc(1)=zero
     do ir=2,ishp
       shapefunc(ir)=shapefunc2(mesh%rad(ir))*mesh%rad(ir)**ll
     end do
   end if
 else if (pshead%shape_type==3) then
   call shapebes(alpha,qq,ll,pshead%rc_hat)
   do ir=1,ishp
     call jbessel(jbes1,besp,bespp,ll,0,qq(1)*mesh%rad(ir))
     call jbessel(jbes2,besp,bespp,ll,0,qq(2)*mesh%rad(ir))
     shapefunc(ir)=shapefunc3(jbes1,jbes2)
   end do
 end if

 if (ishp<mesh%mesh_size) shapefunc(ishp+1:mesh%mesh_size)=zero

!Shape function normalization
 if (pshead%shape_type==-1.or.pshead%shape_type==1.or.pshead%shape_type==2) then
   ALLOCATE(r2k(mesh%mesh_size))
   r2k=zero
   r2k(2:ishp)=shapefunc(2:ishp)*mesh%rad(2:ishp)**(2+ll)
   call simp_gen(norm,r2k,mesh,r_for_intg=pshead%rc_hat)
   norm=one/norm
!   call simp_gen(norm,r2k,mesh);norm=one/norm
   shapefunc(1:ishp)=shapefunc(1:ishp)*norm
   if (pshead%shape_type==-1) norm=one
   DEALLOCATE(r2k)
 else if (pshead%shape_type==3) then
   norm=one
 end if

 end subroutine pawshpfun


!================================================================================
! ROUTINE: FIND PARAMETERS OF BESSEL SHAPE FUNCTION
!================================================================================

!! FUNCTION
!!    Find al and ql parameters for a "Bessel" shape function:
!!    Shape(r)=al1.jl(ql1.r)+al2.jl(ql2.r)
!!      such as Shape(r) and 2 derivatives are zero at r=rc
!!              Intg_0_rc[Shape(r).r^(l+2).dr]=1
!!
!! INPUTS
!!  ll= l quantum number
!!  rc= cut-off radius
!!
!! OUTPUT
!!  al(2)= al coefficients
!!  ql(2)= ql factors

 subroutine shapebes(al,ql,ll,rc)

!Arguments
 integer :: ll
 real(dp) :: rc
 real(dp) :: al(2),ql(2)

!Local variables
 integer :: ii
 real(dp) :: alpha,beta,det,jbes,jbesp,jbespp,qr
 real(dp) :: amat(2,2),bb(2)

!------------------------------------------------------------------

 alpha=1._dp;beta=0._dp
 call solvbes(ql,alpha,beta,ll,2)
 ql(1:2)=ql(1:2)/rc

 do ii=1,2
   qr=ql(ii)*rc
   call jbessel(jbes,jbesp,jbespp,ll,1,qr)
   amat(1,ii)=jbesp*ql(ii)
   call jbessel(jbes,jbesp,jbespp,ll+1,0,qr)
   amat(2,ii)=jbes*rc**(ll+2)/ql(ii)  !  Intg_0_rc[jl(qr).r^(l+2).dr]
 end do

 bb(1)=zero;bb(2)=one

 det=amat(1,1)*amat(2,2)-amat(1,2)*amat(2,1)
 al(1)=(amat(2,2)*bb(1)-amat(1,2)*bb(2))/det
 al(2)=(amat(1,1)*bb(2)-amat(2,1)*bb(1))/det

 end subroutine shapebes


!================================================================================
! ROUTINE: INITIALIZE ARRAYS IN A RADMESH DATASTRUCTURE
!================================================================================

!! FUNCTION
!! Compute all points of a radial mesh
!! Grid can be regular or logarithimc
!!
!! INPUTS
!! mesh=data containing radial grid information
!! r_for_intg=upper boundary for (future) integration over the radial grid
!!            (can be negative for an integration over the whole grid)
!!
!! SIDE EFFECTS
!!  mesh <type(pawrad_type)>=data containing radial grid information
!!  The following quantities are calculated inside the routine:
!!    %stepint = Radial step used to convert any function from the
!!               present grid onto a regular grid in order to integrate it using trapeze method
!!    %rad(mesh_size) = Coordinates of all the points of the mesh.
!!    %radfact(mesh_size) = Factors used to compute radial integrals.
!!    %int_meshsz = Integrals will be computed up to r(int_meshsz)
!!    %simfact(mesh_size) = Factor used to compute radial integrals by the a Simpson scheme
!!                          Integral[f] = Sum_i [simfact(i)*f(i)]
!!    %rmax = Max. value of r = rad(mesh_size)
!!
!! NOTES
!!  Possible mesh types (mesh%mesh_type)
!!   mesh_type=1 (regular grid): rad(i)=(i-1)*AA
!!   mesh_type=2 (logari. grid): rad(i)=AA*(exp[BB*(i-1)]-1)
!!   mesh_type=3 (logari. grid): rad(i>1)=AA*exp[BB*(i-2)] and rad(1)=0
!!   mesh_type=4 (logari. grid): rad(i)=-AA*ln[1-BB*(i-1)] with BB=1/n
!!   mesh_type=5 ( grid): rad(i)=AA*i/(n-i) 

 subroutine init_radmesh(mesh,r_for_intg)

!Arguments
 real(dp),intent(in) :: r_for_intg
 type(pawrad_type),intent(inout) :: mesh

!Local variables
 integer :: ir,ir_last,isim
 real(dp) :: hh

!------------------------------------------------------------------

 allocate(mesh%rad(mesh%mesh_size))
 allocate(mesh%radfact(mesh%mesh_size))
 allocate(mesh%simfact(mesh%mesh_size))

 if (mesh%mesh_type==1) then
   isim=3
   mesh%stepint=mesh%rstep
   mesh%rad(1)=zero;mesh%radfact(1)=one
   do ir=2,mesh%mesh_size
     mesh%rad(ir) =mesh%rstep*dble(ir-1)
     mesh%radfact(ir)=one
   end do
 else if (mesh%mesh_type==2) then
   isim=3
   mesh%stepint=mesh%lstep
   mesh%rad(1)=zero;mesh%radfact(1)=mesh%rstep
   do ir=2,mesh%mesh_size
     mesh%rad(ir) =mesh%rstep*(exp(mesh%lstep*dble(ir-1))-one)
     mesh%radfact(ir)=mesh%rad(ir)+mesh%rstep
   end do
 else if (mesh%mesh_type==3) then
   isim=4
   mesh%stepint=mesh%lstep
   mesh%rad(1)=zero;mesh%radfact(1)=zero
   do ir=2,mesh%mesh_size
     mesh%rad(ir) =mesh%rstep*exp(mesh%lstep*dble(ir-2))
     mesh%radfact(ir)=mesh%rad(ir)
   end do
 else if (mesh%mesh_type==4) then
   isim=3
   mesh%lstep=one/dble(mesh%mesh_size)
   mesh%stepint=mesh%lstep
   mesh%rad(1)=zero;mesh%radfact(1)=mesh%rstep
   do ir=2,mesh%mesh_size
     mesh%rad(ir) =-mesh%rstep*log(one-mesh%lstep*dble(ir-1))
     mesh%radfact(ir)=mesh%rstep/(one-mesh%lstep*dble(ir-1))
   end do
 else if (mesh%mesh_type==5) then
   isim=3
   mesh%stepint=one
   mesh%rad(1)=zero;mesh%radfact(1)=mesh%rstep/mesh%lstep
   do ir=2,mesh%mesh_size
     mesh%rad(ir) =mesh%rstep*dble(ir-1)/(mesh%lstep-dble(ir-1))
     mesh%radfact(ir)=(mesh%rstep+mesh%rad(ir))/(mesh%lstep-dble(ir-1))
   end do

 else !  Other values of mesh_type are not allowed (see psp7in.F90)
  write(std_out,'(a,i0)')" Unknown value of mesh_type: ",mesh%mesh_type
  stop
 end if

 mesh%int_meshsz=mesh%mesh_size
 if (r_for_intg>0.d0) then
   ir=min(ifromr(mesh,r_for_intg),mesh%mesh_size)
   if (ir<mesh%mesh_size) then
     if (abs(mesh%rad(ir+1)-r_for_intg)<abs(mesh%rad(ir)-r_for_intg)) ir=ir+1
   end if
   if (ir>1) then
     if (abs(mesh%rad(ir-1)-r_for_intg)<abs(mesh%rad(ir)-r_for_intg)) ir=ir-1
   end if
   mesh%int_meshsz=ir
 end if

 hh=mesh%stepint/3.d0
 mesh%simfact(mesh%int_meshsz)=hh*mesh%radfact(mesh%int_meshsz)
 mesh%simfact(1:isim-2)=zero
 ir_last=1
 do ir=mesh%int_meshsz,isim,-2
   mesh%simfact(ir-1)=4.d0*hh*mesh%radfact(ir-1)
   mesh%simfact(ir-2)=2.d0*hh*mesh%radfact(ir-2)
   ir_last=ir-2
 end do
 mesh%simfact(ir_last)=half*mesh%simfact(ir_last)
 if (mesh%int_meshsz<mesh%mesh_size) mesh%simfact(mesh%int_meshsz+1:mesh%mesh_size)=zero

 end subroutine init_radmesh


!================================================================================
! ROUTINE: DESTROY ARRAYS IN A RADMESH DATASTRUCTURE
!================================================================================

!! FUNCTION
!! Destroy a radmesh datastructure
!!
!! INPUTS
!! mesh=data containing radial grid information
!!
!! SIDE EFFECTS
!!  mesh <type(pawrad_type)>=data containing radial grid information
!!
 subroutine destroy_radmesh(mesh)

!Arguments
 type(pawrad_type),intent(inout) :: mesh

!Local variables

!------------------------------------------------------------------

 if (allocated(mesh%rad))     deallocate(mesh%rad)
 if (allocated(mesh%radfact)) deallocate(mesh%radfact)
 if (allocated(mesh%simfact)) deallocate(mesh%simfact)
 mesh%mesh_size=0
 mesh%islog=0
 mesh%mesh_type=-1
 mesh%int_meshsz=0
 mesh%rstep=0.d0
 mesh%lstep=0.d0
 mesh%stepint=0.d0

 end subroutine destroy_radmesh


!================================================================================
! ROUTINE: EXTRAPOLATE VALUE AT R=ZERO FOR ANY RADIAL MESH
!================================================================================

!! FUNCTION
!! Extrapolate r=0 value of a function from values near r=0
!! using a 3 points formula
!!
!! INPUTS
!!  funcsz=size of array func
!!  radmesh <type(pawrad_type)>=data containing radial grid information
!!
!! SIDE EFFECTS
!!  func(funcsz)=array containing values of function to extrapolate
!!

 subroutine deducer0(func,funcsz,radmesh)

!Arguments
 integer,intent(in) :: funcsz
 type(pawrad_type),intent(in) :: radmesh
 real(dp),intent(inout) :: func(funcsz)

!------------------------------------------------------------------

 if (radmesh%mesh_type==1.or.radmesh%mesh_type==2.or.radmesh%mesh_type==4.or.radmesh%mesh_type==5) then
   func(1)=func(4)+3*(func(2)-func(3))
 else if (radmesh%mesh_type==3) then
   func(1)=func(4)+exp(two*radmesh%lstep)/(exp(radmesh%lstep)-one)*(func(2)-func(3))
 end if

 end subroutine deducer0


!================================================================================
! ROUTINE: COMPUTE SIMPSON INTEGRAL
!================================================================================

!! FUNCTION
!! Do integral on a given (generalized) grid using Simpson rule
!!
!! INPUTS
!!  radmesh <type(pawrad_type)>=data containing radial grid information
!!  func(:)=integrand values
!!  r_for_intg=upper boundary for (future) integration over the radial grid
!!            (can be negative for an integration over the whole grid)
!!
!! OUTPUT
!!  intg=resulting integral by Simpson rule
!!
!! NOTES
!!  Possible mesh types (radmesh%mesh_type)
!!   mesh_type=1 (regular grid): rad(i)=(i-1)*AA
!!   mesh_type=2 (logari. grid): rad(i)=AA*(exp[BB*(i-1)]-1)
!!   mesh_type=3 (logari. grid): rad(i>1)=AA*exp[BB*(i-2)] and rad(1)=0
!!   mesh_type=4 (logari. grid): rad(i)=-AA*ln[1-BB*(i-1)] with BB=1/n
!!   mesh_type=5 ( grid): rad(i)=AA*i/(n-i)

 subroutine simp_gen(intg,func,radmesh,r_for_intg)

!Arguments
 real(dp),intent(out) :: intg
 real(dp),intent(in),optional :: r_for_intg
 type(pawrad_type),intent(in) :: radmesh
 real(dp),intent(in) :: func(radmesh%int_meshsz)

!Local variables
 integer :: ii,int_meshsz,ir,ir_last,isim,nn
 real(dp) :: hh,resid,simp
 real(dp),allocatable :: simfact(:)

!------------------------------------------------------------------

 if (present(r_for_intg)) then 
   if (r_for_intg>0.d0) then
     ir=min(ifromr(radmesh,r_for_intg),radmesh%mesh_size)
     if (ir<radmesh%mesh_size) then
       if (abs(radmesh%rad(ir+1)-r_for_intg)<abs(radmesh%rad(ir)-r_for_intg)) ir=ir+1
     end if
     if (ir>1) then
       if (abs(radmesh%rad(ir-1)-r_for_intg)<abs(radmesh%rad(ir)-r_for_intg)) ir=ir-1
     end if
     int_meshsz=ir
   end if
   if (int_meshsz>radmesh%int_meshsz) then
     write(std_out,'(a,i4,a,i4)')"simp_gen: BUG int_meshsz= ",int_meshsz," > radmesh%int_meshsz= ",radmesh%int_meshsz
     stop
   end if
   isim=3; if (radmesh%mesh_type==3)isim=4
   ALLOCATE(simfact(radmesh%mesh_size))
   hh=radmesh%stepint/3.d0
   simfact(int_meshsz)=hh*radmesh%radfact(int_meshsz)
   simfact(1:isim-2)=zero
   ir_last=1
   do ir=int_meshsz,isim,-2
     simfact(ir-1)=4.d0*hh*radmesh%radfact(ir-1)
     simfact(ir-2)=2.d0*hh*radmesh%radfact(ir-2)
     ir_last=ir-2
   end do
   simfact(ir_last)=half*simfact(ir_last)
   if (int_meshsz<radmesh%mesh_size) simfact(int_meshsz+1:radmesh%mesh_size)=zero
 
   nn=int_meshsz
   simp=zero
   do ii=1,nn
     simp=simp+func(ii)*simfact(ii)
   end do
   DEALLOCATE(simfact)
 else
   nn=radmesh%int_meshsz

   simp=zero
   do ii=1,nn
     simp=simp+func(ii)*radmesh%simfact(ii)
   end do
 end if

 resid=zero
 if (radmesh%mesh_type==3) then
   resid=half*(func(2)+func(1))*(radmesh%rad(2)-radmesh%rad(1))
   if (mod(nn,2)==1) resid=resid+radmesh%stepint/3.d0*(1.25d0*func(2)*radmesh%radfact(2) &
&   +2.d0*func(3)*radmesh%radfact(3)-0.25d0*func(4)*radmesh%radfact(4))
 else if (mod(nn,2)==0) then
   resid=radmesh%stepint/3.d0*(1.25d0*func(1)*radmesh%radfact(1)+2.d0*func(2)*radmesh%radfact(2) &
&   -0.25d0*func(3)*radmesh%radfact(3))
 end if

 intg=simp+resid

 end subroutine simp_gen


!================================================================================
! ROUTINE: SOLVE POISSON EQUATION
!================================================================================

!! FUNCTION
!!  Solve poisson equation for angularly dependent charge
!!  distribution of angular momentum l
!!  Densities and potentials are given on a (generalized) radial grid
!!
!! INPUTS
!!  den(radmesh%mesh_size)= electron density * (4*pi*r**2) appropriate for l
!!  ll= l quantum number
!!  radmesh <type(pawrad_type)>=data containing radial grid information
!!
!! OUTPUT
!!  qq= lth moment of the charge
!!  rv(radmesh%mesh_size)= electrostatic potential * r in (Hartree*Bohr) units
!!          where v(r)=\frac{1}{2l+1}(\frac{int[(r''^(l+2))g(r'')dr'']} {r^(l+1)}

subroutine poisson(den,ll,qq,radmesh,rv)

!Arguments
 integer,intent(in) :: ll
 real(dp),intent(out) :: qq
 type(pawrad_type),intent(in) :: radmesh
 real(dp),intent(in) :: den(radmesh%mesh_size)
 real(dp),intent(out) :: rv(radmesh%mesh_size)

!Local variables
 integer :: ir,jr,mm,nn
 real(dp) :: angm,hh,rr
 real(dp) :: aa(radmesh%mesh_size),bb(radmesh%mesh_size),cc(radmesh%mesh_size)
 real(dp) :: dd(radmesh%mesh_size+1),ee(4)
 real(dp),allocatable :: radl(:),radl1(:)

!------------------------------------------------------------------

 if (radmesh%mesh_type==1) then

!----------------------------------
!  UNIFORM GRID - NUMEROV ALGORITHM

   hh=radmesh%rstep
   nn=radmesh%int_meshsz-1
   do ir=1,nn
     aa(ir)=two*hh*den(ir+1)/(ir)
     bb(ir)=den(ir+1)*((ir*hh)**ll)
   end do
   dd(1)=zero
   dd(2:nn+1)=bb(1:nn)
   call simp_gen(qq,dd,radmesh)
   qq=qq/(2*ll+1)
   rv(1)=aa(1)+0.1_dp*aa(2)
   do ir=2,nn-1
     rv(ir)=aa(ir)+0.1_dp*(aa(ir+1)+aa(ir-1))
   end do
   rv(nn)=aa(nn)+0.1_dp*aa(nn-1)
   angm=dble(ll*(ll+1))
   rr=(nn+1)*hh
   rv(nn)=rv(nn)+(2.4_dp-0.2_dp*angm/((nn+1)**2))*qq/(rr**ll)
   do ir=1,nn
     aa(ir)=angm/(ir*ir)
     bb(ir)=2.4_dp+aa(ir)
   end do
   do ir=1,nn-1
     cc(ir)=-1.2_dp+0.1_dp*aa(ir+1)
   end do
   do ir=nn,2,-1
     aa(ir)=-1.2_dp+0.1_dp*aa(ir-1)
   end do
   if (nn.eq.1) then
     rv(2)=rv(1)/bb(1)
     rv(1)=zero
   else
     do ir=2,nn
       bb(ir)=bb(ir)-aa(ir)*cc(ir-1)/bb(ir-1)
     end do
     rv(1)=rv(1)/bb(1)
     do ir=2,nn
       rv(ir)=(rv(ir)-aa(ir)*rv(ir-1))/bb(ir)
     end do
     do ir=nn-1,1,-1
       rv(ir)=rv(ir)-cc(ir)*rv(ir+1)/bb(ir)
     end do
     do ir=nn+1,2,-1
       rv(ir)=rv(ir-1)
     end do
     rv(1)=zero
     if (nn+1<radmesh%mesh_size) rv(nn+2:radmesh%mesh_size)=zero
   end if
   rv(:)=half*rv(:)

 else

!----------------------------------
!  ANY OTHER GRID - SIMPSON ALGORITHM

   nn=radmesh%mesh_size;hh=third*radmesh%stepint
   do while (abs(den(nn))<tol16.and.nn>radmesh%int_meshsz)
     nn=nn-1
   end do
    mm=nn;if (radmesh%mesh_type==3) mm=mm-1
   ALLOCATE(radl(nn))
   ALLOCATE(radl1(nn))
   do jr=nn,2,-1
     ir=nn-jr+1
     radl(jr) =radmesh%rad(jr)**ll
     radl1(jr)=radmesh%rad(jr)*radl(jr)
     aa(ir)=den(jr)*radmesh%radfact(jr)*radl(jr)
     bb(ir)=den(jr)*radmesh%radfact(jr)/radl1(jr)
   end do
   radl(1)=zero;radl1(1)=zero
   ee(2)=aa(nn-1);ee(3)=aa(nn-2);ee(4)=aa(nn-3)
   call deducer0(ee,4,radmesh)
   aa(nn)=ee(1)
   ee(2)=bb(nn-1);ee(3)=bb(nn-2);ee(4)=bb(nn-3)
   call deducer0(ee,4,radmesh)
   bb(nn)=ee(1)
   cc(1)=zero;dd(1)=zero
   do ir=3,mm,2
     cc(ir)  =cc(ir-2)+hh*(aa(ir-2)+four*aa(ir-1)+aa(ir))
     cc(ir-1)=cc(ir-2)+hh*(1.25_dp*aa(ir-2)+two*aa(ir-1)-quarter*aa(ir))
     dd(ir)  =dd(ir-2)+hh*(bb(ir-2)+four*bb(ir-1)+bb(ir))
     dd(ir-1)=dd(ir-2)+hh*(1.25_dp*bb(ir-2)+two*bb(ir-1)-quarter*bb(ir))
   end do
   if (mod(mm,2)==0) then
     cc(mm)=cc(mm-2)+hh*(aa(mm-2)+four*aa(mm-1)+aa(mm))
     dd(mm)=dd(mm-2)+hh*(bb(mm-2)+four*bb(mm-1)+bb(mm))
   end if
   if (mm<nn) then
     cc(nn)=cc(mm)+half*(aa(mm)+aa(nn))*(radmesh%rad(1+nn-mm)-radmesh%rad(1))
     dd(nn)=dd(mm)+half*(bb(mm)+bb(nn))*(radmesh%rad(1+nn-mm)-radmesh%rad(1))
   end if
   rv(1)=zero
   do ir=2,nn
     jr=nn-ir+1
     rv(ir)=(dd(jr)*radl1(ir)+(cc(nn)-cc(jr))/radl(ir))/(two*ll+one)
   end do
   if (nn<radmesh%mesh_size) rv(nn+1:radmesh%mesh_size)=rv(nn)
   qq=cc(nn)/(two*ll+one)
   DEALLOCATE(radl)
   DEALLOCATE(radl1)
 end if

 end subroutine poisson


!================================================================================
! FUNCTION: GET AN INDEX IN A RADIAL MESH
!================================================================================

!! FUNCTION
!! Retreive Index FROM a given R value in a radial grid
!! Grid can be regular or logarithimc
!!
!! INPUTS
!!  rr=given input r value
!!  radmesh <type(pawrad_type)>=data containing radial grid information
!!
!! OUTPUT
!!  ifromr=index of rr in radial grid
!!
!! NOTES
!!  Possible mesh types (radmesh%mesh_type)
!!   mesh_type=1 (regular grid): rad(i)=(i-1)*AA
!!   mesh_type=2 (logari. grid): rad(i)=AA*(exp[BB*(i-1)]-1)
!!   mesh_type=3 (logari. grid): rad(i>1)=AA*exp[BB*(i-2)] and rad(1)=0
!!   mesh_type=4 (logari. grid): rad(i)=-AA*ln[1-BB*(i-1)] with BB=1/n
!!   mesh_type=5 ( grid): rad(i)=AA*i/(n-i)

 integer function ifromr(radmesh,rr)

!Arguments
 real(dp),intent(in) :: rr
 type(pawrad_type),intent(in) :: radmesh

!Local variables

!------------------------------------------------------------------

 if (radmesh%mesh_type==1) then
   ifromr=int(tol8+rr/radmesh%rstep)+1
 else if (radmesh%mesh_type==2) then
   ifromr=int(tol8+log(1.d0+rr/radmesh%rstep)/radmesh%lstep)+1
 else if (radmesh%mesh_type==3) then
   if (rr<radmesh%rstep) then
     ifromr=1
   else
     ifromr=int(tol8+log(rr/radmesh%rstep)/radmesh%lstep)+2
   end if
 else if (radmesh%mesh_type==4) then
   ifromr=int(tol8+(1.d0-exp(-rr/radmesh%rstep))/radmesh%lstep)+1
 else if (radmesh%mesh_type==5) then
   ifromr=int(tol8+(radmesh%lstep*rr)/(radmesh%rstep+rr))+1
 else
!  Other values of mesh_type are not allowed (see psp7in.F90)
   write(std_out,'(a,i0)')" Unknown value of %mesh_type ",radmesh%mesh_type
   stop
 end if

 end function ifromr


!================================================================================
! ROUTINE: COMPUTE SECOND DERIVATIVE OF A FUNCTION FROM A CUBIC SPLINE
!================================================================================

!! FUNCTION
!!  SPLINE (originally SPLINE_CUBIC_SET) computes the second derivatives
!!  of a cubic spline.
!!
!! INPUTS
!!    Input, integer N, the number of data points; N must be at least 2. 
!!    In the special case where N = 2 and IBCBEG = IBCEND = 0, the 
!!    spline will actually be linear. 
!!
!!    Input, double precision T(N), the knot values, that is, the points where data
!!    is specified.  The knot values should be distinct, and increasing.
!!
!!    Input, double precision Y(N), the data values to be interpolated.
!!
!!    Input, integer IBCBEG, left boundary condition flag:
!!      0: the spline should be a quadratic over the first interval;
!!      1: the first derivative at the left endpoint should be YBCBEG;
!!      2: the second derivative at the left endpoint should be YBCBEG.
!!    Input, integer IBCEND, right boundary condition flag:
!!      0: the spline should be a quadratic over the last interval;
!!      1: the first derivative at the right endpoint should be YBCEND;
!!      2: the second derivative at the right endpoint should be YBCEND.
!!
!! OUTPUT
!!    Output, double precision YPP(N), the second derivatives of the cubic spline.
!!    Work space, double precision DIAG(N) - should be removed ...
!!
!! NOTES
!!    From numerical recipies...

 subroutine spline( t, y, n, ybcbeg, ybcend, ypp )

!Arguments
  integer, intent(in) :: n
  double precision, intent(in) :: t(n)
  double precision, intent(in) :: y(n)
  double precision, intent(in) :: ybcbeg
  double precision, intent(in) :: ybcend
  double precision, intent(out) :: ypp(n)

!Local variables
  integer :: ibcbeg
  integer :: ibcend
  integer :: i,k
  double precision :: ratio,pinv
  double precision, allocatable :: tmp(:)

!------------------------------------------------------------------

!  Check.
  if ( n <= 1 ) then
    write(std_out,* ) ' '
    write(std_out,* ) 'SPLINE_CUBIC_SET - Fatal error!'
    write(std_out,* ) '  The number of knots must be at least 2.'
    write(std_out,* ) '  The input value of N = ', n
    write(std_out,* ) "Fatal error"
    stop
  end if

  ALLOCATE(tmp(n))
  do i = 1, n-1
    if ( t(i) >= t(i+1) ) then
      write(std_out,* ) ' '
      write(std_out,* ) 'SPLINE_CUBIC_SET - Fatal error!'
      write(std_out,* ) '  The knots must be strictly increasing, but'
      write(std_out,* ) '  T(',  i,') = ', t(i)
      write(std_out,* ) '  T(',i+1,') = ', t(i+1)
      write(std_out,* ) "Fatal error"
      stop
    end if
  end do

  ibcbeg=1 ; ibcend=1
  if(ybcbeg>1.0d+30)ibcbeg=0
  if(ybcend>1.0d+30)ibcend=0

! Set the first and last equations.
  if ( ibcbeg == 0 ) then
    ypp(1) = 0.d0
    tmp(1) = 0.d0
  else if ( ibcbeg == 1 ) then
    ypp(1) = -0.5d0
    tmp(1) = (3.d0/(t(2)-t(1)))*((y(2)-y(1))/(t(2)-t(1))-ybcbeg)
  end if
  if ( ibcend == 0 ) then
    ypp(n) = 0.d0
    tmp(n) = 0.d0
  else if ( ibcend == 1 ) then
    ypp(n) = 0.5d0
    tmp(n) = (3.d0/(t(n)-t(n-1)))*(ybcend-(y(n)-y(n-1))/(t(n)-t(n-1)))
  end if

! Set the intermediate equations.
  do i=2,n-1
   ratio=(t(i)-t(i-1))/(t(i+1)-t(i-1))
   pinv = 1.0d0/(ratio*ypp(i-1) + 2.0d0)
   ypp(i) = (ratio-1.0d0)*pinv
   tmp(i)=(6.0d0*((y(i+1)-y(i))/(t(i+1)-t(i))-(y(i)-y(i-1)) &
&    /(t(i)-t(i-1)))/(t(i+1)-t(i-1))-ratio*tmp(i-1))*pinv
   if (abs(tmp(i))<1.d5*tiny(0.d0)) tmp(i)=0.d0   !MT20050927
  enddo

! Solve the equations
  ypp(n) = (tmp(n)-ypp(n)*tmp(n-1))/(ypp(n)*ypp(n-1)+1.0d0)
  do k=n-1,1,-1
   ypp(k)=ypp(k)*ypp(k+1)+tmp(k)
  enddo

  DEALLOCATE(tmp)

 end subroutine spline


!================================================================================
! ROUTINE: DO A CUBIC SPLINE
!================================================================================

!! FUNCTION
!!  Compute spline interpolation. There is no hypothesis
!!  about the spacing of the input grid points.
!!
!! INPUTS
!!  nspline: number of grid points of input mesh
!!  xspline(nspline): input mesh
!!  yspline(nspline): function on input mesh
!!  ysplin2(nspline): second derivative of yspline on input mesh
!!  nfit: number of points of output mesh
!!  xfit(nfit): output mesh
!!
!! OUTPUT
!!  yfit(nfit): function on output mesh
!!  [ierr]=A non-zero value is used to signal that some points in xfit exceed xspline(nspline).
!!    The input value is incremented by the number of such points.

 subroutine splint(nspline,xspline,yspline,ysplin2,nfit,xfit,yfit,ierr)

!Arguments
 integer, intent(in) :: nfit, nspline
 integer,optional,intent(out) :: ierr
 double precision, intent(in) :: xspline(nspline)
 double precision, intent(in) :: yspline(nspline)
 double precision, intent(in) :: ysplin2(nspline)
 double precision, intent(in) :: xfit(nfit)
 double precision, intent(out) :: yfit(nfit)

!Local variables
 integer :: left,i,k,right,my_err
 double precision :: delarg,invdelarg,aa,bb

!------------------------------------------------------------------

 my_err=0 ; left = 1
 do i=1, nfit
   yfit(i)=0.d0  ! Initialize for the unlikely event that rmax exceed r(mesh)
   do k=left+1, nspline
     if(xspline(k) >= xfit(i)) then
       if(xspline(k-1) <= xfit(i)) then
         right = k
         left = k-1
       else
         if (k-1.eq.1 .and. i.eq.1) then
           write(std_out,* )'xfit(1) < xspline(1)'
           stop
         else
           write(std_out,* )'xfit not properly ordered'
           stop
         end if
       end if
       delarg= xspline(right) - xspline(left)
       invdelarg= 1.0d0/delarg
       aa= (xspline(right)-xfit(i))*invdelarg
       bb= (xfit(i)-xspline(left))*invdelarg

       yfit(i) = aa*yspline(left) + bb*yspline(right)    &
&               +( (aa*aa*aa-aa)*ysplin2(left) +         &
&                  (bb*bb*bb-bb)*ysplin2(right) ) *delarg*delarg/6.0d0
       exit
     end if
   end do ! k
   if (k==nspline+1) my_err=my_err+1 ! xfit not found 
 end do ! i
 if (PRESENT(ierr)) ierr=my_err

 end subroutine splint


!================================================================================
! ROUTINE: COMPUTE DERIVATIVES OF A FUnCTION AT THE BOUNDARIES OF AN INTERVAL
!================================================================================

!! FUNCTION
!! Computes derivatives of a function a boundaries of interval (first and last derivative)
!!
!! INPUTS
!!  func(n)= array containing function
!!  mesh <type(pawrad_type)>= radial mesh and related data
!!  nn= size of intervall
!!
!! OUTPUT
!!  yp1,ypn= derivatives of func at r(1) and r(n)
!!

 subroutine bound_deriv(func,mesh,nn,yp1,ypn)

!Arguments
 integer, intent(in) :: nn
 real(dp), intent(in) :: func(nn)
 real(dp), intent(out) :: yp1,ypn
 type(pawrad_type),intent(in) :: mesh

!------------------------------------------------------------------

 if (mesh%radfact(1)>zero) then
   yp1=1._dp/12._dp/mesh%stepint/mesh%radfact(1) &
&   *(-25._dp*func(1)+48._dp*func(2)-36._dp*func(3)+16._dp*func(4)-3._dp*func(5))
 else
   yp1=(func(2)-func(1))/(mesh%rad(2)-mesh%rad(1))
 end if
 ypn=1._dp/12._dp/mesh%stepint &
& *( 3._dp*func(nn-4)-16._dp*func(nn-3)+36._dp*func(nn-2)-48._dp*func(nn-1) &
& +25._dp*func(nn))/mesh%radfact(nn)

 end subroutine bound_deriv


!================================================================================
! ROUTINE: FIND ROOTS OF INTRINSIC EQUATION: alpha.jl(Q) + beta.Q.djl/dr(Q)=0
!================================================================================

!! FUNCTION
!!    Find nq first roots of instrinsic equation:
!!               alpha.jl(Q) + beta.Q.djl/dr(Q) = 0
!!
!! INPUTS
!!  alpha,beta= factors in intrinsic equation
!!  ll= l quantum number
!!  nq= number of roots to find
!!
!! OUTPUT
!!  root(nq)= roots of instrinsic equation

 subroutine solvbes(root,alpha,beta,ll,nq)

!Arguments
 integer :: ll,nq
 real(dp) :: alpha,beta
 real(dp) :: root(nq)

!Local variables
 integer :: nroot
 real(dp),parameter :: dh=0.1_dp,tol=tol14
 real(dp) :: dum,hh,jbes,jbesp,qq,qx,y1,y2

!------------------------------------------------------------------

 qq=dh;nroot=0

 do while (nroot<nq)
   call jbessel(jbes,jbesp,dum,ll,1,qq)
   y1=alpha*jbes+beta*qq*jbesp
   qq=qq+dh
   call jbessel(jbes,jbesp,dum,ll,1,qq)
   y2=alpha*jbes+beta*qq*jbesp

   do while (y1*y2>=zero)
     qq=qq+dh
     call jbessel(jbes,jbesp,dum,ll,1,qq)
     y2=alpha*jbes+beta*qq*jbesp
   end do

   hh=dh;qx=qq
   do while (hh>tol)
     hh=half*hh
     if (y1*y2<zero) then
       qx=qx-hh
     else
       qx=qx+hh
     end if
     call jbessel(jbes,jbesp,dum,ll,1,qx)
     y2=alpha*jbes+beta*qx*jbesp
   end do
   nroot=nroot+1
   root(nroot)=qx
 end do

 end subroutine solvbes


!================================================================================
! ROUTINE: COMPUTE SPHERICAL BESSEL FUNCTION VALUES
!================================================================================

!! FUNCTION
!! Compute spherical Bessel function j_l(x) and derivative(s)
!!
!! INPUTS
!!  ll=l-order of the Bessel function
!!  order=1 if first derivative is requested
!!        2 if first and second derivatives are requested
!!  xx=where to compute j_l
!!
!! OUTPUT
!!  bes= Bessel function j_l at xx
!!  besp= first derivative of j_l at xx (only if order>=1)
!!  bespp= second derivative of j_l at xx (only if order=2)

 subroutine jbessel(bes,besp,bespp,ll,order,xx)

!Arguments
 integer,intent(in) :: ll,order
 real(dp),intent(in) :: xx
 real(dp),intent(out) :: bes,besp,bespp

!Local variables
 integer,parameter :: imax=40
 integer :: ii,il
 real(dp),parameter :: prec=1.d-15
 real(dp) :: besp1,fact,factp,factpp,jn,jnp,jnpp,jr,xx2,xxinv

!------------------------------------------------------------------

 if (order>2) then 
   write(std_out,* )"Wrong order in jbessel!"
   stop
 end if

 if (abs(xx)<prec) then
   bes=zero;if (ll==0) bes=one
   if (order>=1) then
     besp=zero;if (ll==1) besp=third
   end if
   if (order==2) then
     bespp=zero
     if (ll==0) bespp=-third
     if (ll==2) bespp=2.d0/15.d0
   end if
   return
 end if

 xxinv=one/xx
 if (order==0) then
   factp=zero
   factpp=zero
   jnp=zero
   jnpp=zero
 end if 

 if (xx<one) then
   xx2=0.5d0*xx*xx
   fact=one
   do il=1,ll
     fact=fact*xx/dble(2*il+1)
   end do
   jn=one;jr=one;ii=0
   do while(abs(jr)>=prec.and.ii<imax)
     ii=ii+1;jr=-jr*xx2/dble(ii*(2*(ll+ii)+1))
     jn=jn+jr
   end do
   bes=jn*fact
   if (abs(jr)>prec) then 
     write(std_out,* )'Bessel function did not converge!'
     stop
   end if
   if (order>=1) then
     factp=fact*xx/dble(2*ll+3)
     jnp=one;jr=one;ii=0
     do while(abs(jr)>=prec.AND.ii<imax)
       ii=ii+1;jr=-jr*xx2/dble(ii*(2*(ll+ii)+3))
       jnp=jnp+jr
     end do
     besp=-jnp*factp+jn*fact*xxinv*dble(ll)
     if (abs(jr)>prec) then 
       write(std_out,* )'1st der. of Bessel function did not converge!'
       stop
     end if
   end if
   if (order==2) then
     factpp=factp*xx/dble(2*ll+5)
     jnpp=one;jr=one;ii=0
     do while(abs(jr)>=prec.AND.ii<imax)
       ii=ii+1;jr=-jr*xx2/dble(ii*(2*(ll+ii)+5))
       jnpp=jnpp+jr
     end do
     besp1=-jnpp*factpp+jnp*factp*xxinv*dble(ll+1)
     if (abs(jr)>prec) then 
       write(std_out,* )'2nd der. of Bessel function did not converge !'
       stop
     end if
   end if
 else
   jn =sin(xx)*xxinv
   jnp=(-cos(xx)+jn)*xxinv
   do il=2,ll+1
     jr=-jn+dble(2*il-1)*jnp*xxinv
     jn=jnp;jnp=jr
   end do
   bes=jn
   if (order>=1) besp =-jnp+jn *xxinv*dble(ll)
   if (order==2) besp1= jn -jnp*xxinv*dble(ll+2)
 end if

 if (order==2) bespp=-besp1+besp*ll*xxinv-bes*ll*xxinv*xxinv

end subroutine jbessel


!================================================================================
! ROUTINE: CONVERT AN INTEGER INTO A STRING
!================================================================================

!! FUNCTION
!!  Take an integer (<=4 digits) and return it
!!   in the form of a character string
!!
!! INPUTS
!!  i=input integer
!!
!! OUTPUT
!!  stuff=output string

SUBROUTINE mkname(ii,stuff)

!Arguments
 INTEGER :: ii
 CHARACTER(4) :: stuff

!Local variables
 INTEGER :: i1,i10,i100,i1000

!------------------------------------------------------------------

 stuff='?'
 IF (ii.GT.10000) ii=MOD(ii,10000)
 i1000=ii/1000
 i100=(ii-1000*i1000)/100
 i10=(ii-1000*i1000-100*i100)/10
 i1=(ii-1000*i1000-100*i100-10*i10)
 IF (ii>=1000) THEN
    stuff=CHAR(i1000+48)//CHAR(i100+48)//CHAR(i10+48)//CHAR(i1+48)
    RETURN
 ENDIF
 IF (ii>=100) THEN
    stuff=CHAR(i100+48)//CHAR(i10+48)//CHAR(i1+48)
    RETURN
 ENDIF
 IF (ii>=10) THEN
    stuff=CHAR(i10+48)//CHAR(i1+48)
    RETURN
 ENDIF
 IF (ii>=0) stuff=CHAR(i1+48)

 END SUBROUTINE mkname


!================================================================================
! FUNCTION: STRIP A STRING
!================================================================================

!! FUNCTION
!!  Strip a string
!!
!! SIDE EFFECTS
!!  inputchar=string to be processed

 CHARACTER(len=fnlen) function stripchar(inputchar)

!Local variables
 INTEGER :: ii,jj,nn
 CHARACTER*(*), INTENT(IN) :: inputchar

!------------------------------------------------------------------

 nn=LEN(inputchar)
 do ii=1,fnlen
   stripchar(ii:ii)=''
 enddo
 jj=0
 do ii=1,nn
   if (inputchar(ii:ii) /= '') then
      jj=jj+1
      stripchar(jj:jj)=inputchar(ii:ii)
   endif
 enddo

 end function stripchar


!================================================================================
! END
!================================================================================

 end program
